# mypy: disable-error-code=attr-defined
#
# HELP: Disable mypy check(attr-defined) since there is no good way to
# typing a Mixin class like ItemViewNoScroll: https://github.com/python/typing/issues/213
#
# TODO(cosven): Such mixin class(like ItemViewNoScrollMixin) has little reable and
# it's hard to typing. I think we can split it into a delegate and a mixin to solve
# this problem.
#
# For example::
#
#     class ItemViewNoScrollManager: ...
#     class ItemViewNoScrollMixin: ...
#     class ListView(ItemViewNoScrollMixin, QListView):
#         def __init__(self):
#             self.no_scroll_manager = ItemViewNoScrollManager(self)
#     class NoScrollableItemView(Protocol):
#         no_scroll_manager: ItemViewNoScrollManager

from __future__ import annotations
import asyncio
import random
import sys
import logging
from contextlib import contextmanager
from typing import TypeVar, List, Optional, Generic, Union, cast, TYPE_CHECKING

from PyQt5.QtCore import QModelIndex, QSize, Qt, pyqtSignal, QSortFilterProxyModel, \
    QAbstractListModel, QPoint
from PyQt5.QtGui import QPalette, QFontMetrics, QColor, QPainter, QMouseEvent, \
    QKeySequence
from PyQt5.QtWidgets import QApplication, QScrollArea, QWidget, QShortcut

from feeluown.utils.aio import run_afn, run_fn
from feeluown.utils.reader import AsyncReader, Reader
from feeluown.utils.typing_ import Protocol
from feeluown.excs import ProviderIOError, ResourceNotFound
from feeluown.library import ModelNotFound, ModelType, BaseModel
from feeluown.library import reverse
from feeluown.gui.consts import FontFamilies


if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


logger = logging.getLogger(__name__)
T = TypeVar("T")

IS_MACOS = sys.platform == 'darwin'


def darker_or_lighter(color: QColor, factor):
    """
    If the color is a light color, for example, white, then this returns
    a darker color.
    """
    if color.lightness() > 150:
        return color.darker(factor)
    return color.lighter(factor)


async def async_run(func, loop=None, executor=None):
    """异步的获取 model 属性值

    值得注意的是，如果 executor 消费队列排队了，这个会卡住。
    """
    if loop is None:
        loop = asyncio.get_event_loop()
    return await loop.run_in_executor(executor, func)


class ActionError(Exception):
    pass


def get_qapp() -> QApplication:
    return cast(QApplication, QApplication.instance())


def is_macos():
    """Check if operating system is macOS

    .. versionadded: v3.7.10
    """
    return sys.platform == 'darwin'


def disconnect_slots_if_has(signal):
    try:
        signal.disconnect()
    except TypeError:  # signal has no slots
        pass


def resize_font(font, delta):
    if font.pixelSize() > 0:
        font.setPixelSize(max(1, font.pixelSize() + delta))
    else:
        font.setPointSize(max(1, font.pointSize() + delta))


def palette_set_bg_color(palette, color):
    """set palette background color"""
    if sys.platform == 'linux':
        # KDE use the QPalette.Base as background color
        palette.setColor(QPalette.Active, QPalette.Base, color)
        palette.setColor(QPalette.Inactive, QPalette.Base, color)
        # GNOME use the QPalette.Window as background color
        palette.setColor(QPalette.Active, QPalette.Window, color)
        palette.setColor(QPalette.Inactive, QPalette.Window, color)
    else:
        # macOS use the QPalette.Window as background color
        palette.setColor(QPalette.Active, QPalette.Window, color)
        palette.setColor(QPalette.Inactive, QPalette.Window, color)


class BgTransparentMixin:
    def __init__(self: QWidget, *args, **kwargs):  # type: ignore[misc]
        palette = self.palette()
        palette_set_bg_color(palette, Qt.transparent)
        self.setPalette(palette)


class BaseScrollAreaForNoScrollItemView(QScrollArea):
    """A scroll area base class for itemview with no_scroll_v=True

    Subclass must implement the following methods:
    * itemview
    * height_for_itemview
    Also note the API is not stable.

    .. versionadded:: 3.8.9
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.verticalScrollBar().valueChanged.connect(self.on_v_scrollbar_value_changed)

    def get_itemview(self):
        raise NotImplementedError

    def height_for_itemview(self):
        raise NotImplementedError

    def resizeEvent(self, e):
        super().resizeEvent(e)
        self.maybe_resize_itemview()

    def maybe_resize_itemview(self):
        """Resize itemview to make sure it has a chance to fetch more items

        When the itemview has no more items, do not need to resize it.
        """
        itemview = self.get_itemview()
        if itemview is not None:
            model = itemview.model()
            if model is not None and model.canFetchMore(QModelIndex()):
                # +1 to make sure user can scroll, then user has a change to
                # trigger itemview fetch more.
                itemview.suggest_min_height(self.height_for_itemview() + 1)
                itemview.adjust_height()

    def maybe_trigger_itemview_fetch_more(self):
        itemview = self.get_itemview()
        if itemview is not None:
            model = itemview.model()
            if model is not None and model.canFetchMore(QModelIndex()):
                model.fetchMore(QModelIndex())

    def on_v_scrollbar_value_changed(self, value):
        maximum = self.verticalScrollBar().maximum()
        if maximum == value:
            self.maybe_trigger_itemview_fetch_more()


class ItemViewNoScrollMixin:
    """
    `no_scroll_v` means that the itemview's size(height) is always enough to hold
    all fetched items. When new items are fetched, the itemview size is
    automatically adjueted.

    The itemview with no_scroll_v=True is usually used with an outside ScrollArea.

    Python Notes::

        ItemViewNoScrollMixin follows the "cooperative multi-inheritance" pattern.
        Since ItemViewNoScrollMixin use some QObject API, subclass should consider
        the MRO order. In other words, ItemViewNoScrollMixin should be the parent
        class of QObject. XWidget(ItemViewNoScrollMixin, QObject) is a good choice,
        and XWidget(QObject, ItemViewNoScrollMixin) is not.
    """
    def __init__(self, *args, no_scroll_v=True, row_height=0, least_row_count=0,
                 fixed_row_count=0, reserved=30, **kwargs):
        """
        :params no_scroll_v: enable on no_scroll_v feature or not
        :params fixed_row_count: set row_height when fixed_row_count is set

        .. versionadded:: 3.7.8
           The *row_height*, *least_row_count*, *reserved* parameter were added.

        .. versionadded:: 3.8.9
           The *fixed_row_count* parameter was added.
        """
        super().__init__(**kwargs)  # Cooperative multi-inheritance.
        self._least_row_count = least_row_count
        self._fixed_row_count = fixed_row_count
        self._row_height = row_height
        self._reserved = reserved

        self._min_height = 0

        self._no_scroll_v = no_scroll_v

    def initialize(self):
        """
        .. versionadded:: 3.7.7
        """
        if self._no_scroll_v is True:
            self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

    # def set_no_scroll_v(self, no_scroll_v):
    #     """
    #     .. versionadded:: 3.7.7
    #     """
    #     self._no_scroll_v = no_scroll_v
    #     if no_scroll_v is True:
    #         self.adjust_height()
    #         self.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOff)

    def adjust_height(self):
        if self.model() is None:
            return

        if self.model().canFetchMore(QModelIndex()):
            if self._fixed_row_count == 0:
                # according to QAbstractItemView source code,
                # qt will trigger fetchMore when the last row is
                # inside the viewport, so we always hide the last
                # two row to ensure fetch-more will not be
                # triggered automatically.
                #
                # qt triggers fetchMore when user scrolls down to bottom.
                index = self._last_visible_index()
                rect = self.visualRect(index)
                height = self.sizeHint().height() - int(rect.height() * 1.5) - \
                    self._reserved
                self.setFixedHeight(max(height, self.min_height()))
            else:
                self.setFixedHeight(self._row_height * self._fixed_row_count)
        else:
            height = self.sizeHint().height()
            self.setFixedHeight(height)
        self.updateGeometry()

    def on_rows_changed(self, *args):
        if self._no_scroll_v is True:
            self.adjust_height()

    def setModel(self, model):
        super().setModel(model)  # type: ignore[misc]
        if model is None:
            return
        model.rowsInserted.connect(self.on_rows_changed)
        model.rowsRemoved.connect(self.on_rows_changed)
        if isinstance(model, QSortFilterProxyModel):
            srcModel = model.sourceModel()
            if isinstance(srcModel, ReaderFetchMoreMixin):
                srcModel.no_more_item.connect(self.on_rows_changed)
        self.on_rows_changed()

    def wheelEvent(self, e):
        if self._no_scroll_v is True:
            if abs(e.angleDelta().x()) > abs(e.angleDelta().y()):
                QApplication.sendEvent(self.horizontalScrollBar(), e)
            else:
                e.ignore()  # let parents handle it
        else:
            super().wheelEvent(e)  # type: ignore[misc]

    def sizeHint(self):
        super_size_hint = super().sizeHint()  # type: ignore[misc]
        if self._no_scroll_v is False:
            return super_size_hint

        height = min_height = self.min_height()
        if self.model() is not None:
            if self._fixed_row_count == 0:
                index = self._last_visible_index()
                rect = self.visualRect(index)
                height = rect.y() + rect.height() + self._reserved
                height = max(min_height, height)
            else:
                height = self._row_height * self._fixed_row_count
        return QSize(super_size_hint.width(), height)

    def _last_visible_index(self):
        source_model = self.model()
        row_index = source_model.rowCount() - 1
        if isinstance(source_model, QAbstractListModel):
            column_index = 0
        else:
            column_index = source_model.columnCount() - 1
            while column_index >= 0:
                index = source_model.index(row_index, column_index)
                if not self.isIndexHidden(index):
                    break
                column_index -= 1
        # can't use createIndex here, why?
        return source_model.index(row_index, column_index)

    def min_height(self):
        default = self._row_height * self._least_row_count + self._reserved
        return max(self._min_height, default)

    def suggest_min_height(self, height):
        """
        parent should call this method where it's size changed
        """
        self._min_height = height


def elided_text(text, width, font=None):
    if font is None:
        font = QApplication.font()
    font_metrics = QFontMetrics(font)
    return font_metrics.elidedText(text, Qt.ElideRight, width)


class Paddings(tuple):
    def __new__(cls, *sequence):
        assert len(sequence) == 4
        return super().__new__(cls, tuple(sequence))

    @property
    def left(self):
        return self[0]

    @property
    def top(self):
        return self[1]

    @property
    def right(self):
        return self[2]

    @property
    def bottom(self):
        return self[3]

    @property
    def height(self):
        return self[1] + self[3]

    @property
    def width(self):
        return self[0] + self[2]


Margins = Paddings


# HELP(cosven): please help remove duplicate code between ReaderFetchMoreMixin
# and this class.
class ModelUsingReader(Protocol[T]):
    _reader: Union[Reader[T], AsyncReader]
    _items: List[T]
    _fetch_more_step: int
    _is_fetching: bool

    def rowCount(self) -> int: ...
    def beginInsertRows(self, _, __, ___): ...
    def endInsertRows(self): ...
    def canFetchMore(self, _) -> bool: ...
    def fetchMore(self, _): ...
    def can_fetch_more(self, _=None) -> bool: ...
    def fetch_more_impl(self): ...
    def on_items_fetched(self, items: List[T]): ...
    def _fetch_more_cb(self, items: Optional[List[T]]): ...
    def _async_fetch_cb(self, future): ...


class ReaderFetchMoreMixin(Generic[T]):
    """
    The class should implement

    1. _reader
    2. _items
    3. _fetch_more_step
    4. _is_fetching
    """
    no_more_item = pyqtSignal()

    def canFetchMore(self: ModelUsingReader[T], _):
        return self.can_fetch_more()

    def fetchMore(self: ModelUsingReader[T], _):
        if self._is_fetching is False:
            self._is_fetching = True
            self.fetch_more_impl()

    def can_fetch_more(self: ModelUsingReader[T], _=None):
        reader = cast(Reader, self._reader)
        count = reader.count
        if count is not None:
            return count > self.rowCount()

        # The reader sets the count when it has no more items,
        # so it is safe to return True here
        return True

    def fetch_more_impl(self: ModelUsingReader[T]):
        """fetch more items from reader
        """
        reader = self._reader
        step = self._fetch_more_step

        if isinstance(reader, AsyncReader):
            async def fetch():
                items = []
                count = 0
                async for item in reader:  # type: ignore
                    items.append(item)
                    count += 1
                    if count == step:
                        break
                return items
            task = run_afn(fetch)
        else:
            assert isinstance(reader, Reader)
            task = run_fn(reader.read_range, self.rowCount(), step + self.rowCount())
        task.add_done_callback(self._async_fetch_cb)

    def on_items_fetched(self: ModelUsingReader[T], items: List[T]):
        begin = len(self._items)
        end = begin + len(items) - 1
        self.beginInsertRows(QModelIndex(), begin, end)
        self._items.extend(items)
        self.endInsertRows()

    def _fetch_more_cb(self: ModelUsingReader[T], items: Optional[List[T]]):
        self._is_fetching = False
        if items is not None:
            self.on_items_fetched(items)

    def _async_fetch_cb(self: ModelUsingReader[T], future):
        try:
            items = future.result()
        except ProviderIOError as e:
            logger.error(f'async fetch more items failed, reason: {e}')
            self._fetch_more_cb(None)
        except:  # noqa
            logger.exception('async fetch more items failed')
            self._fetch_more_cb(None)
        else:
            if not items:
                # The reader should not return empty list when fetching more items,
                # maybe something wrong with the reader.
                logger.warning('async fetch more items return empty list')
            self._fetch_more_cb(items)


def fetch_cover_wrapper(app: GuiApp):
    """
    Your should only use this helper within ImgListModel and SongMiniCardListModel.
    """
    img_mgr, library = app.img_mgr, app.library

    async def fetch_image_with_cb(img_uid, img_url, cb):
        # Fetch image by url and invoke cb.
        if img_url:
            # FIXME: sleep random second to avoid send too many request to provider
            await asyncio.sleep(random.randrange(100) / 100)
            content = await img_mgr.get(img_url, img_uid)
            cb(content)
        else:
            cb(None)

    async def fetch_song_pic_from_album(album, cb):
        if album is not None:
            await fetch_other_model_cover(album, cb)
        else:
            cb(None)

    async def fetch_song_pic_url(model, cb):
        """
        The song may be a v2 brief model or a v1 model.
        """
        is_v2_model = isinstance(model, BaseModel)

        # If the song is a v1 model, just fallback to use its album cover.
        if not is_v2_model:
            try:
                upgraded_song = await run_fn(library.song_upgrade, model)
            except ModelNotFound:
                cb(None)
            else:
                await fetch_song_pic_from_album(upgraded_song.album, cb)
            return

        # v2 song model has its own image(pic_url), check if it is in cache first.
        cache_key = 'album_cover_uid'
        song_img_uid = reverse(model) + '/pic_url'
        album_img_uid, _ = model.cache_get(cache_key)
        for img_uid in (song_img_uid, album_img_uid):
            if img_uid:
                content = img_mgr.get_from_cache(img_uid)
                if content is not None:
                    cb(content)
                    return

        # Image is not in cache.
        try:
            upgraded_song = await run_fn(library.song_upgrade, model)
        except ResourceNotFound:
            cb(None)
        else:
            # Try to fetch with pic_url first.
            # Note that some providers may not provide pic_url for songs.
            if upgraded_song.pic_url:
                img_uid = reverse(model) + '/pic_url'
                img_url = upgraded_song.pic_url
                return await fetch_image_with_cb(img_uid, img_url, cb)

            album = upgraded_song.album
            if album is None:
                cb(None)
                return

            album_img_uid = reverse(album) + '/cover'
            model.cache_set(cache_key, album_img_uid)
            return await fetch_song_pic_from_album(album, cb)

    async def fetch_other_model_cover(model, cb):
        img_uid = reverse(model) + '/cover'

        # Check image cache with image unique ID.
        content = img_mgr.get_from_cache(img_uid)
        if content is not None:
            cb(content)
            return

        img_url = await run_fn(library.model_get_cover, model)
        return await fetch_image_with_cb(img_uid, img_url, cb)

    async def fetch_model_cover(model, cb):
        if ModelType(model.meta.model_type) is ModelType.song:
            return await fetch_song_pic_url(model, cb)
        return await fetch_other_model_cover(model, cb)

    return fetch_model_cover


def random_solarized_color():
    return QColor(random.choice(list(SOLARIZED_COLORS.values())))


@contextmanager
def painter_save(painter: QPainter):
    painter.save()
    yield
    painter.restore()


def secondary_text_color(palette: QPalette):
    text_color: QColor = palette.color(QPalette.Text)
    if text_color.lightness() > 150:
        non_text_color = text_color.darker(140)
    else:
        non_text_color = text_color.lighter(150)
    non_text_color.setAlpha(100)
    return non_text_color


class ClickableMixin:
    clicked = pyqtSignal()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)  # Cooperative multi-inheritance.

        self._down = False

    def mousePressEvent(self, e: QMouseEvent):
        if e.button() != Qt.LeftButton:
            # Call super.mousePressEvent because the concrete class may do sth inside it.
            super().mousePressEvent(e)
            return
        if self._hit_button(e.pos()):
            self._down = True
            e.accept()
        else:
            super().mousePressEvent(e)

    def mouseReleaseEvent(self, e: QMouseEvent):
        if e.button() != Qt.LeftButton or not self._down:
            super().mouseReleaseEvent(e)
            return
        self._down = False
        if self._hit_button(e.pos()):
            self.clicked.emit()
            e.accept()
        else:
            super().mouseReleaseEvent(e)

    def _hit_button(self, pos: QPoint):
        return self.rect().contains(pos)

    def set_down(self, down: bool):
        self._down = down


def set_default_font_families(widget):
    font = widget.font()
    font.setFamilies(FontFamilies)
    widget.setFont(font)


def esc_hide_widget(widget):
    QShortcut(QKeySequence.Cancel, widget).activated.connect(widget.hide)


# https://ethanschoonover.com/solarized/
SOLARIZED_COLORS = {
    'yellow':    '#b58900',
    'orange':    '#cb4b16',
    'red':       '#dc322f',
    'magenta':   '#d33682',
    'violet':    '#6c71c4',
    'blue':      '#268bd2',
    'cyan':      '#2aa198',
    'green':     '#859900',
}



import logging

logger = logging.getLogger(__name__)


def run_mpris2_server(app):
    # pylint: disable=import-outside-toplevel
    try:
        import dbus
        import dbus.service
        import dbus.mainloop.pyqt5
    except ImportError as e:
        logger.error("can't run mpris2 server: %s",  str(e))
    else:
        from .mpris2 import Mpris2Service, BusName

        # check if a mainloop was already set
        mainloop = dbus.get_default_main_loop()
        if mainloop is not None:
            logger.warning("mpris2 service already enabled? "
                           "maybe you should remove feeluown-mpris2-plugin")
            return

        # set the mainloop before any dbus operation
        dbus.mainloop.pyqt5.DBusQtMainLoop(set_as_default=True)
        session_bus = dbus.SessionBus()
        bus = dbus.service.BusName(BusName, session_bus)
        service = Mpris2Service(app, bus)
        service.enable()



# -*- coding: utf-8 -*-

from __future__ import annotations
import logging
import json
import os
import sys
from collections import defaultdict
from typing import TYPE_CHECKING

from PyQt5.QtCore import pyqtSignal, QObject
from PyQt5.QtGui import QGuiApplication, QPalette, QColor
from feeluown.utils.utils import get_osx_theme
from feeluown.gui.helpers import get_qapp

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp

logger = logging.getLogger(__name__)


Dark = DARK = 'dark'
Light = LIGHT = 'light'
MacOSDark = 'macos_dark'  # for Linux/Windows Users

Roles = ['AlternateBase', 'Background', 'Base', 'BrightText',
         'Button', 'ButtonText', 'Dark', 'Foreground', 'Highlight',
         'HighlightedText', 'Light', 'Link', 'LinkVisited', 'Mid',
         'Midlight', 'NoRole', 'PlaceholderText', 'Shadow', 'Text',
         'ToolTipBase', 'ToolTipText', 'Window', 'WindowText']

Groups = ['Disabled', 'Active', 'Inactive']


def read_resource(filename):
    filepath = os.path.abspath(__file__)
    dirname = os.path.dirname(filepath)
    qssfilepath = os.path.join(dirname, 'assets/themes/{}'.format(filename))
    with open(qssfilepath, encoding='UTF-8') as f:
        s = f.read()
    return s


class ThemeManager(QObject):
    """colors/icons manager"""

    theme_changed = pyqtSignal(str)

    def __init__(self, app: GuiApp, parent=None):
        super().__init__(parent=parent)
        self._app = app
        self.theme = None

    def initialize(self):
        # XXX: I don't know why we should autoload twice
        # to make it work well on Linux(GNOME)
        self.autoload()
        self._app.started.connect(lambda app: self.autoload(), weak=False)
        get_qapp().paletteChanged.connect(lambda p: self.autoload())

    def autoload(self):
        if self._app.config.THEME == 'auto':
            if sys.platform == 'darwin':
                if get_osx_theme() == 1:
                    theme = DARK
                else:
                    theme = LIGHT
            else:
                theme = self.guess_system_theme()
                if theme == DARK:
                    theme = Dark
        else:  # user settings have highest priority
            theme = self._app.config.THEME
        self.load_theme(theme)

    def load_theme(self, theme):
        if theme == DARK:
            self.load_dark()
        elif theme == MacOSDark:
            self.load_macos_dark()
        else:
            self.load_light()
        self.theme = theme
        self.theme_changed.emit(theme)

    def guess_system_theme(self):
        palette = self._app.palette()
        bg_color = palette.color(QPalette.Window)
        if bg_color.lightness() > 150:
            return LIGHT
        return DARK

    def load_light(self):
        common = read_resource('common.qss')
        light = read_resource('light.qss')
        get_qapp().setStyleSheet(common + light)

    def load_dark(self):
        common = read_resource('common.qss')
        dark = read_resource('dark.qss')
        get_qapp().setStyleSheet(common + dark)

    def load_macos_dark(self):
        """
        So many DEs on Linux! Hard to compat! We give them macOS
        dark theme colors!

        Users can also design a theme colors by themselves,
        we provider dump_colors/load_colors function for conviniece.
        """
        self.load_dark()
        content = read_resource('macos_dark.colors')
        colors = json.loads(content)
        try:
            get_qapp().paletteChanged.disconnect(self.autoload)
        except TypeError:
            pass
        palette = load_colors(colors)
        self._app.setPalette(palette)
        get_qapp().paletteChanged.connect(self.autoload)

    def get_pressed_color(self):
        """pressed color for button-like widget

        In Feeluown, we have two kinds of buttons, text and icon. All TextButtons
        have same color in pressed state.

        THINKING: we can create a API for theme manager like QPalette.color if we need
        """
        if self.guess_system_theme() == DARK:
            return '#3e3e3e'
        return '#DDD'

    def get_icon(self, name):
        """get icon by name

        this API is similar to QIcon.fromTheme
        """
        theme_kind = self.guess_system_theme()
        if name == 'tray':
            if theme_kind == DARK:
                return 'icons:tray-dark.png'
            return 'icons:tray-light.png'


def dump_colors():
    json_ = defaultdict(dict)  # type: ignore[var-annotated]
    palette = QGuiApplication.palette()
    for group_attr in Groups:
        group = getattr(QPalette, group_attr)
        for role_attr in Roles:
            role = getattr(QPalette, role_attr)
            json_[group_attr][role_attr] = palette.color(group, role).name()
    return json_


def load_colors(colors):
    palette = QGuiApplication.palette()
    for group_attr, value in colors.items():
        for role_attr, color_name in value.items():
            try:
                role = getattr(QPalette, role_attr)
                group = getattr(QPalette, group_attr)
                palette.setColor(group, role, QColor(color_name))
            except AttributeError:
                pass
    QGuiApplication.setPalette(palette)
    return palette



import logging
from typing import TYPE_CHECKING

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QFrame, QHBoxLayout, QVBoxLayout, QPushButton, QSizePolicy
from feeluown.gui.widgets.cover_label import CoverLabelV2

from feeluown.utils.aio import run_afn
from feeluown.gui.widgets.volume_button import VolumeButton
from feeluown.gui.widgets.progress_slider import ProgressSlider
from feeluown.gui.widgets.labels import ProgressLabel, DurationLabel
from feeluown.gui.components import (
    LineSongLabel, MediaButtonsV2, LyricButton, WatchButton, LikeButton,
    NowplayingMVTextButton, PlaylistButton, SongSourceTag,
)
from feeluown.gui.helpers import IS_MACOS, ClickableMixin

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp

logger = logging.getLogger(__name__)


class ClickableCover(ClickableMixin, CoverLabelV2):
    def __init__(self, app, **kwargs):
        super().__init__(app=app, **kwargs)


class PlayerControlPanel(QFrame):

    def __init__(self, app: 'GuiApp', parent=None):
        super().__init__(parent)
        self._app = app

        class IconButton(QPushButton):
            def __init__(self, *args, **kwargs):
                super().__init__(*args, **kwargs)

        _button_width = 36

        # initialize sub widgets
        self._layout = QHBoxLayout(self)

        self.media_btns = MediaButtonsV2(
            app=self._app, spacing=8, button_width=_button_width, parent=self
        )
        self.previous_btn = self.media_btns.previous_btn
        self.pp_btn = self.media_btns.pp_btn
        self.next_btn = self.media_btns.next_btn
        self.volume_btn = VolumeButton(length=_button_width, parent=self)
        self.playlist_btn = PlaylistButton(self._app, length=_button_width, parent=self)
        #: mark song as favorite button
        self.like_btn = LikeButton(self._app, parent=self)
        self.mv_btn = NowplayingMVTextButton(app=self._app, parent=self)
        self.toggle_lyric_btn = LyricButton(self._app, parent=self)
        self.download_btn = QPushButton(self)
        self.toggle_watch_btn = WatchButton(self._app, self)

        self.playlist_btn.setObjectName('playlist_btn')
        self.volume_btn.setObjectName('volume_btn')
        self.download_btn.setObjectName('download_btn')

        self.progress_slider = ProgressSlider(app=app, parent=self)

        self.volume_btn.setToolTip('调整音量')
        self.playlist_btn.setToolTip('显示当前播放列表')
        self.download_btn.setToolTip('下载歌曲（未实现，欢迎 PR）')
        self.download_btn.setCheckable(True)

        self.song_title_label = LineSongLabel(self._app)
        self.song_title_label.setAlignment(Qt.AlignCenter)
        self.song_source_label = SongSourceTag(self._app, parent=self)

        self.cover_label = ClickableCover(app)
        self.duration_label = DurationLabel(app, parent=self)
        self.position_label = ProgressLabel(app, parent=self)

        # we should enable focus since we want to have shortcut keys
        self.setFocusPolicy(Qt.StrongFocus)

        self.volume_btn.change_volume_needed.connect(
            lambda volume: setattr(self._app.player, 'volume', volume))
        self._app.player.video_format_changed.connect(
            self.on_video_format_changed, aioqueue=True)
        player = self._app.player
        player.metadata_changed.connect(self.on_metadata_changed, aioqueue=True)
        player.volume_changed.connect(self.volume_btn.on_volume_changed, aioqueue=True)
        self.cover_label.clicked.connect(self.show_nowplaying_overlay)

        self._setup_ui()

    def _setup_ui(self):
        self.cover_label.setFixedWidth(44)
        self.cover_label.setMaximumHeight(44)
        self.song_source_label.setFixedHeight(20)
        if IS_MACOS:
            self.progress_slider.setFixedHeight(25)  # half of parent height
        else:
            self.progress_slider.setFixedHeight(20)  # half of parent height
        self.position_label.setFixedWidth(50)
        self.duration_label.setFixedWidth(50)
        self.position_label.setAlignment(Qt.AlignCenter)
        self.duration_label.setAlignment(Qt.AlignCenter)
        self.download_btn.setFixedSize(15, 15)
        self.download_btn.hide()
        self.mv_btn.setFixedHeight(16)
        self.toggle_watch_btn.setFixedHeight(16)

        self.progress_slider.setSizePolicy(QSizePolicy.Expanding,
                                           QSizePolicy.Preferred)
        self._sub_layout = QVBoxLayout()
        self._sub_top_layout = QHBoxLayout()
        self._progress_v_layout = QVBoxLayout()
        self._cover_v_layout = QVBoxLayout()

        # add space to make top layout align with progress slider
        self._sub_top_layout.addSpacing(3)
        self._sub_top_layout.addWidget(self.song_source_label)
        self._sub_top_layout.addSpacing(5)
        self._sub_top_layout.addWidget(self.song_title_label)
        self._sub_top_layout.addSpacing(5)
        self._sub_top_layout.addWidget(self.like_btn)
        self._sub_top_layout.addSpacing(8)
        self._sub_top_layout.addWidget(self.mv_btn)
        self._sub_top_layout.addSpacing(8)
        self._sub_top_layout.addWidget(self.toggle_lyric_btn)
        self._sub_top_layout.addSpacing(8)
        self._sub_top_layout.addWidget(self.toggle_watch_btn)
        # self._sub_top_layout.addSpacing(8)
        # self._sub_top_layout.addWidget(self.download_btn)
        self._sub_top_layout.addSpacing(3)

        self._sub_layout.addSpacing(3)
        self._sub_layout.addStretch(0)
        self._sub_layout.addLayout(self._sub_top_layout)
        self._sub_layout.addStretch(0)
        self._sub_layout.addWidget(self.progress_slider)
        self._sub_layout.addStretch(0)

        self._progress_v_layout.addStretch(0)
        self._progress_v_layout.addWidget(self.position_label)
        self._progress_v_layout.addSpacing(3)
        self._progress_v_layout.addWidget(self.duration_label)
        self._progress_v_layout.addStretch(0)

        # Put the cover_label in a vboxlayout and add strech around it,
        # so that cover_label's sizehint is respected.
        self._cover_v_layout.addStretch(0)
        self._cover_v_layout.addWidget(self.cover_label)
        self._cover_v_layout.addStretch(0)

        self._layout.addSpacing(20)
        self._layout.addWidget(self.media_btns)
        self._layout.addSpacing(26)
        self._layout.addWidget(self.volume_btn)
        # 18 = 200(left_panel_width) - 4 * 30(btn) - 20 - 8 - 8 -26
        self._layout.addSpacing(50)
        self._layout.addSpacing(18)
        self._layout.addStretch(0)
        self._layout.addLayout(self._cover_v_layout)
        self._layout.addSpacing(7)
        self._layout.addLayout(self._sub_layout)
        self._layout.setStretchFactor(self._sub_layout, 1)
        self._layout.addSpacing(7)
        self._layout.addLayout(self._progress_v_layout)
        self._layout.addStretch(0)
        self._layout.addSpacing(18)
        self._layout.addWidget(self.playlist_btn)
        self._layout.addSpacing(18)
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)

    def on_metadata_changed(self, metadata):
        metadata = metadata or {}
        released = metadata.get('released', '')
        if released:
            self.cover_label.setToolTip(f'专辑发行日期：{released}')
        else:
            self.cover_label.setToolTip('')
        # Set song artwork.
        artwork = metadata.get('artwork', '')
        artwork_uid = metadata.get('uri', artwork)
        if artwork:
            run_afn(self.cover_label.show_cover, artwork, artwork_uid)
        else:
            self.cover_label.show_img(None)

    def on_video_format_changed(self, video_format):
        if video_format is None:
            self.media_btns.toggle_video_btn.hide()
        else:
            self.media_btns.toggle_video_btn.show()

    def show_nowplaying_overlay(self):
        self._app.ui.nowplaying_overlay.show()
        self._app.ui.nowplaying_overlay.raise_()


class TopPanel(QFrame):
    def __init__(self, app, parent=None):
        super().__init__(parent)
        self._app = app

        self._layout = QHBoxLayout(self)
        self.pc_panel = PlayerControlPanel(self._app, self)
        self.setObjectName('top_panel')

        self.setFixedHeight(60)

        self._layout.addWidget(self.pc_panel)



import logging
import inspect
import warnings
from collections import deque
from urllib.parse import urlencode
from typing import Optional, Deque

from feeluown.utils import aio
from feeluown.utils.router import Router, NotFound
from feeluown.library import resolve, reverse, ResolveFailed, parse_line

logger = logging.getLogger(__name__)


MODEL_PAGE_PREFIX = '/models/'


class Browser:
    """GUI 页面管理中心

    将 feeluown 类比为浏览器：magicbox 是搜索框，RightPanel
    部分是浏览器主体。
    """

    def __init__(self, app):
        self._app = app
        # The two stack are used to save history
        # TODO: Currently, browser only save the whole page path in history, and the
        # related data such as model is not cached. Caching the data should improve
        # performance.
        self._back_stack: Deque[str] = deque(maxlen=10)
        self._forward_stack: Deque[str] = deque(maxlen=10)
        self.router = Router()  # alpha

        self._last_page: Optional[str] = None
        self.current_page: Optional[str] = None

        #: the value in local_storage must be string,
        # please follow the convention
        self.local_storage = {}

    @property
    def ui(self):
        return self._app.ui

    def goto(self, model=None, path=None, page=None, query=None, uri=None):
        """Goto page

        Relations between different goto function, and `_goto` is
        the real renderer::

                 -> _goto_model_page -> _goto
            goto -> _goto_page -> _goto_model_page -> _goto
                               -> _goto

        Typical usage::

            goto(model=model, path=path, query=xxx)
            goto(page=page, query=xxx)

        Wrong usage::

            goto(path=page, query=xxx)
        """
        # backward compact: old code use goto(uri=page)
        if uri is not None:
            warnings.warn('please use parameter page', DeprecationWarning)
            page = page or uri

        qs = urlencode(query) if query else ''

        try:
            if model is not None:
                if qs:
                    path = (path or '') + '?' + qs
                self._goto_model_page(model, path)
            else:
                # backward compat: old code use goto(path=page) wrongly
                if path is not None:
                    warnings.warn('please use parameter page')
                page = page or path
                if qs:
                    page = page + '?' + qs
                self._goto_page(page)
        except NotFound:
            logger.warning(f'{page} renderer not found')
        else:
            # save history records
            if self._last_page is not None and self._last_page != self.current_page:
                self._back_stack.append(self._last_page)
            self._forward_stack.clear()
            self.on_history_changed()

    def back(self):
        try:
            page = self._back_stack.pop()
        except IndexError:
            logger.warning("Can't go back.")
        else:
            self._goto_page(page=page)
            assert self._last_page is not None
            self._forward_stack.append(self._last_page)
            self.on_history_changed()

    def forward(self):
        try:
            page = self._forward_stack.pop()
        except IndexError:
            logger.warning("Can't go forward.")
        else:
            self._goto_page(page=page)
            assert self._last_page is not None
            self._back_stack.append(self._last_page)
            self.on_history_changed()

    def route(self, rule):
        """路由装饰器 (alpha)"""
        return self.router.route(rule)

    def _goto_page(self, page):
        # see if the page match the two special cases
        if page.startswith(MODEL_PAGE_PREFIX):
            try:
                # FIXME: resolve is temporarily too magic
                uri = 'fuo://' + page[len(MODEL_PAGE_PREFIX):]
                model, path = parse_line(uri)
                model = resolve(reverse(model))
            except ResolveFailed:
                model = None
                logger.warning(f'invalid model page:{page}')
            else:
                return self._goto_model_page(model, path)
        else:
            self._goto(page, {'app': self._app})

    def _goto_model_page(self, model, path=''):
        """goto model page

        The main difference between model page and other pages is that model page
        has different context. It's context has an extra key `model`.
        """
        path = path or ''
        page = base_page = MODEL_PAGE_PREFIX + reverse(model)[6:]
        if path:
            page = f'{base_page}{path}'
        self._goto(page, {'app': self._app, 'model': model})

    def _goto(self, page, ctx):
        # Do some initialization
        self.ui.toolbar.clear_stacked_widget()
        # To keep backward compat, we add magicbox by default
        self.ui.toolbar.add_stacked_widget(self.ui.magicbox)

        x = self.router.dispatch(page, ctx)
        if inspect.iscoroutine(x):
            aio.create_task(x)

        self._last_page = self.current_page
        self.current_page = page

    @property
    def can_back(self):
        return len(self._back_stack) > 0

    @property
    def can_forward(self):
        return len(self._forward_stack) > 0

    # --------------
    # UI Controllers
    # --------------

    def _render_coll(self, _, identifier):
        coll = self._app.coll_mgr.get(identifier)
        self._app.ui.right_panel.show_collection(coll)

    def on_history_changed(self):
        self.ui.back_btn.setEnabled(self.can_back)
        self.ui.forward_btn.setEnabled(self.can_forward)

    # --------------
    # initialization
    # --------------

    def initialize(self):
        """browser should be initialized after all ui components are created

        1. bind routes with renderer
        """
        from feeluown.gui.pages.search import render as render_search
        from feeluown.gui.pages.model import render as render_model
        from feeluown.gui.pages.coll_mixed import render as render_coll_mixed
        from feeluown.gui.pages.song_explore import render as render_song_explore
        from feeluown.gui.pages.provider_home import render as render_provider_home
        from feeluown.gui.pages.recently_played import render as render_recently_played
        from feeluown.gui.pages.recommendation import render as render_rec
        from feeluown.gui.pages.recommendation_daily_songs import \
            render as render_rec_daily_songs
        from feeluown.gui.pages.my_fav import render as render_my_fav

        model_prefix = f'{MODEL_PAGE_PREFIX}<provider>'

        async def dummy_render(req, *args, **kwargs):
            warnings.warn(f'This route:{req.rule} will be removed.')

        urlpatterns = [
            (f'{model_prefix}/<ns>/<identifier>', render_model),
            # These routes will be removed in v3.9
            (f'{model_prefix}/songs/<identifier>/similar', dummy_render),
            (f'{model_prefix}/songs/<identifier>/hot_comments', dummy_render),
            (f'{model_prefix}/songs/<identifier>/explore', render_song_explore),
            ('/colls/<identifier>', render_coll_mixed),
            ('/providers/<identifier>', render_provider_home),
            ('/recently_played', render_recently_played),
            ('/search', render_search),
            ('/rec', render_rec),
            ('/rec/daily_songs', render_rec_daily_songs),
            ('/my_fav', render_my_fav),
        ]
        for url, renderer in urlpatterns:
            self.route(url)(renderer)



from jsonrpc import JSONRPCResponseManager, Dispatcher

from feeluown.fuoexec.fuoexec import fuoexec_get_globals
from feeluown.serializers import serialize


class DynamicDispatcher(Dispatcher):
    def __getitem__(self, key):
        try:
            return self.method_map[key]
        except KeyError:
            method = eval(key, fuoexec_get_globals())
            return method


dispatcher = DynamicDispatcher()


def initialize():
    dispatcher.add_method(lambda: "pong", name="ping")


def handle(data):
    response = JSONRPCResponseManager.handle(data, dispatcher)
    if 'result' in response.data:
        result = serialize('python', response.result)
        response.result = result
    return response.data



from PyQt5.QtCore import Qt, QModelIndex, QItemSelectionModel
from PyQt5.QtWidgets import QMenu, QAbstractItemView

from feeluown.gui.components import SongMenuInitializer
from feeluown.gui.helpers import fetch_cover_wrapper
from feeluown.gui.widgets.song_minicard_list import (
    SongMiniCardListView,
    SongMiniCardListModel,
)
from feeluown.utils.reader import create_reader


class PlayerPlaylistModel(SongMiniCardListModel):
    # FIXME: make this singleton

    def __init__(self, playlist, *args, **kwargs):
        reader = create_reader(playlist.list())
        super().__init__(reader, *args, **kwargs)

        self._playlist = playlist
        self._playlist.songs_added.connect(self.on_songs_added)
        self._playlist.songs_removed.connect(self.on_songs_removed)

    def flags(self, index):
        flags = super().flags(index)
        song = index.data(Qt.UserRole)[0]
        if self._playlist.is_bad(song):
            # Disable bad song.
            flags &= ~Qt.ItemIsEnabled
        return flags

    def on_songs_added(self, index, count):
        self.beginInsertRows(QModelIndex(), index, index + count - 1)
        # Insert from tail to front.
        while count > 0:
            self._items.insert(index, self._playlist[index + count - 1])
            count -= 1
        self.endInsertRows()

    def on_songs_removed(self, index, count):
        self.beginRemoveRows(QModelIndex(), index, index + count - 1)
        while count > 0:
            self._items.pop(index + count - 1)
            count -= 1
        self.endRemoveRows()


class PlayerPlaylistView(SongMiniCardListView):

    def __init__(self, app, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._app = app

        self.play_song_needed.connect(self._app.playlist.play_model)
        self.setModel(
            PlayerPlaylistModel(
                self._app.playlist,
                fetch_cover_wrapper(self._app),
            )
        )

    def contextMenuEvent(self, e):
        index = self.indexAt(e.pos())
        if not index.isValid():
            return

        song = index.data(Qt.UserRole)[0]
        menu = QMenu()
        action = menu.addAction('从播放队列中移除')
        menu.addSeparator()
        SongMenuInitializer(self._app, song).apply(menu)

        action.triggered.connect(lambda: self._app.playlist.remove(song))
        menu.exec_(e.globalPos())

    def scroll_to_current_song(self):
        """Scroll to the current song, and select it to highlight it."""
        current_song = self._app.playlist.current_song
        songs = self._app.playlist.list()
        if current_song is not None:
            model = self.model()
            row = songs.index(current_song)
            index = model.index(row, 0)
            # In order to highlight the current song.
            self.selectionModel().select(index, QItemSelectionModel.SelectCurrent)
            self.scrollTo(index, QAbstractItemView.PositionAtCenter)



import sys
from contextlib import contextmanager

from PyQt5.QtCore import QMetaObject, pyqtSlot, QSize
from PyQt5.QtOpenGL import QGLContext

from mpv import MpvRenderContext, OpenGlCbGetProcAddrFn, _mpv_set_property_string

from feeluown.gui.widgets.video import VideoOpenGLWidget
from feeluown.gui.helpers import IS_MACOS


def get_proc_addr(_, name):
    glctx = QGLContext.currentContext()
    if glctx is None:
        return 0
    addr = int(glctx.getProcAddress(name.decode('utf-8')))
    return addr


class MpvOpenGLWidget(VideoOpenGLWidget):
    """Mpv 视频输出窗口

    销毁时，应该调用 shutdown 方法来释放资源。

    该 Widget 是模仿一些 C++ 程序编写而成（详见项目 research 目录下的例子），
    我们对背后的原理不太理解，目前测试是可以正常工作的。

    目前主要的疑问有：

    - [ ] shutdown 方法期望是用来释放资源的，目前的写法不知是否合理？
          应该在什么时候调用 shutdown 方法？
    """

    def __init__(self, app, parent=None):
        super().__init__(app=app, parent=parent)
        self._app = app
        self.mpv = self._app.player._mpv  # noqa
        self.ctx = None
        self.get_proc_addr_c = OpenGlCbGetProcAddrFn(get_proc_addr)

    def initializeGL(self):
        params = {'get_proc_address': self.get_proc_addr_c}
        self.ctx = MpvRenderContext(self.mpv,
                                    'opengl',
                                    opengl_init_params=params)
        self.ctx.update_cb = self.on_update

    def shutdown(self):
        if self.ctx is not None:
            self.ctx.free()
            self.ctx = None

    def paintGL(self):
        # HELP: It seems that `initializeGL` is called by Qt on
        # old version (<= v5.15.2)
        if self.ctx is None:
            self.initializeGL()
            assert self.ctx is not None
        # compatible with HiDPI display
        ratio = self._app.devicePixelRatio()
        w = int(self.width() * ratio)
        h = int(self.height() * ratio)
        opengl_fbo = {'w': w,
                      'h': h,
                      'fbo': self.defaultFramebufferObject()}
        self.ctx.render(flip_y=True, opengl_fbo=opengl_fbo)

        for gl_painter in self._gl_painters:
            gl_painter.paint(self)

    @pyqtSlot()
    def maybe_update(self):
        if self.window().isMinimized():
            self.makeCurrent()
            self.paintGL()
            self.context().swapBuffers(self.context().surface())
            self.doneCurrent()
        else:
            self.update()

    def on_update(self, ctx=None):
        # maybe_update method should run on the thread that creates the
        # OpenGLContext, which in general is the main thread.
        # QMetaObject.invokeMethod can do this trick.
        QMetaObject.invokeMethod(self, 'maybe_update')

    # NOTE(cosven): heightForwidth does not work (tested inside nowplaying_overlay.py)
    def hasHeightForWidth(self) -> bool:
        return bool(self.mpv.width)

    def heightForWidth(self, width: int) -> int:
        if self.mpv.width:
            return width // self.mpv.width * self.mpv.height
        return super().heightForWidth(width)

    def sizeHint(self):
        if self.mpv.width:
            return QSize(self.mpv.width, self.mpv.height)
        return super().sizeHint()

    @contextmanager
    def change_parent(self):
        # on macOS, changing mpv widget parent cause no side effects.
        # on Linux (wayland), it seems changing mpv widget parent may cause segfault,
        # so do some hack to avoid crash.
        if not IS_MACOS:
            self._before_change_mpv_widget_parent()
        try:
            yield
        finally:
            if not IS_MACOS:
                self._after_change_mpv_widget_parent()

    def _before_change_mpv_widget_parent(self):
        """
        According to Qt docs, reparenting an OpenGLWidget will destory the GL context.
        In mpv widget, it calls _mpv_opengl_cb_uninit_gl. After uninit_gl, mpv can't show
        video anymore because video_out is destroyed.

        See mpv mpv_opengl_cb_uninit_gl implementation for more details.
        """
        _mpv_set_property_string(self.mpv.handle, b'vid', b'no')
        self.hide()
        self.shutdown()

    def _after_change_mpv_widget_parent(self):
        """
        To recover the video show, we should reinit gl and reinit video. gl is
        automatically reinited when the mpv_widget do painting. We should
        manually reinit video.

        NOTE(cosven): After some investigation, I found that the API in mpv to
        reinit video_out(maybe video is almost same as video_out)
        is init_best_video_out. Theoretically, sending 'video-reload' command
        will trigger this API. However, we can't run this command
        in our case and I don't know why. Another way to trigger
        the API is to switch track. Changing vid property just switch the track.

        Inpect mpv init_best_video_out caller for more details. You should see
        mp_switch_track_n is one of the entrypoint.

        Note the GL must be re-initialized properly before this function is called.
        Generally, show parent once can trigger the initialization. Otherwise,
        mpv.vid will not be set to 1, and video can not be shown.
        """
        self.show()
        self.repaint()  # force repaint to trigger re-initialization
        _mpv_set_property_string(self.mpv.handle, b'vid', b'1')
        if not bool(self.mpv.vid):
            print('WARNING: video widget is not reconfigured properly', file=sys.stderr)


# TODO: 实现 MpvEmbeddedWidget
class MpvEmbeddedWidget:
    pass



import random

from feeluown.utils import aio
from feeluown.app import App

# (nickname, name), no order.
# Write your Chinese(or other languages) name if you want.
Contributors = [
    ('cosven', 'Shaowen Yin'),
    ('felixonmars', ''),
    ('PhotonQuantum', ''),
    ('wuliaotc', ''),
    ('cyliuu', ''),
    ('light4', ''),
    ('hjlarry', ''),
    ('sbwtw', ''),
    ('oryxfea', ''),
    ('poly000', ''),
    ('BruceZhang1993', ''),
    ('cposture', ''),
    ('helinb', ''),
    ('Yexiaoxing', ''),
    ('albertofwb', ''),
    ('catsout', ''),
    ('Torxed', ''),
    ('RealXuChe', ''),
    ('rapiz1', ''),
    ('Heersin', ''),
    ('chen-chao', ''),
    ('keter42', ''),
    ('timgates42', ''),
    ('junhan-z', ''),
    ('CareF', ''),
    ('berberman', ''),
    ('SaneBow', ''),
    ('wsyxbcl', ''),
    ('Thearas', ''),
    ('reaink', ''),
    ('wenLiangcan', ''),
    ('leedagee', ''),
    ('hanchanli', ''),
    ('xssss1', ''),
    ('williamherry', ''),
]


class TipsManager:
    """在合适的时候展示一些使用 Tip"""

    tips = [
        '你知道 FeelUOwn 可以配合 osdlyrics 使用吗?',
        '搜索快捷键是 Ctrl + F',
        "在搜索框输入“>>> app.tips_mgr.show_random()”查看更多 Tips",
        '专辑图片上右键可以查看原图哦 ~',
        '可以拖动歌曲来将歌曲添加到歌单呐！',
        '鼠标悬浮或右键常有惊喜 ~',
        '开启 watch 模式一边看 MV，一边工作学习香不香？'
    ]

    def __init__(self, app: App):
        self._app = app
        self._app.started.connect(self.show_random_after_5s)

    def show_random(self):
        if random.randint(0, 10) < 5:
            self._app.show_msg(random.choice(self.tips), timeout=2500)
        else:
            contributor = random.choice(Contributors)
            nickname, name = contributor
            if name:
                user = f'{name} (@{nickname})'
            else:
                user = f'@{nickname}'
            msg = f'感谢 {user} 的贡献 :)'
            self._app.show_msg(msg, timeout=2500)

    def show_random_after_5s(self, *_):
        async def task():
            await aio.sleep(5)
            self.show_random()

        aio.run_afn(task)



from .typename import attach_typename
from .base import Serializer, SerializerMeta, SerializerError, \
    SimpleSerializerMixin
from .model_helpers import ModelSerializerMixin, SongSerializerMixin, \
    ArtistSerializerMixin, AlbumSerializerMixin, PlaylistSerializerMixin, \
    UserSerializerMixin, SearchSerializerMixin, ProviderSerializerMixin


class PythonSerializer(Serializer):
    _mapping = {}

    def __init__(self, **options):
        super().__init__(**options)
        self.opt_level = options.get('level', 0)

    def serialize_items(self, items):
        json_ = {}
        for key, value in items:
            serializer_cls = PythonSerializer.get_serializer_cls(value)
            serializer = serializer_cls(brief=True, level=self.opt_level + 1)
            json_[key] = serializer.serialize(value)
        return json_

    @attach_typename
    def serialize(self, obj):
        return super().serialize(obj)


class ModelSerializer(PythonSerializer, ModelSerializerMixin):

    def __init__(self, **options):
        if options.get('brief') is False and options.get('fetch') is False:
            raise SerializerError(
                "fetch can't be False when brief is False")

        super().__init__(**options)
        self.opt_as_line = options.get('as_line', False)
        self.opt_brief = options.get('brief', True)
        self.opt_fetch = options.get('fetch', False)

        if self.opt_brief is False:
            self.opt_fetch = True

    @attach_typename
    def serialize(self, model):
        dict_ = {}
        for field, value in self._get_items(model):
            serializer_cls = self.get_serializer_cls(value)
            value_dict = serializer_cls(brief=True, fetch=False).serialize(value)
            dict_[field] = value_dict
        return dict_


#######################
# container serializers
#######################


class ListSerializer(PythonSerializer, metaclass=SerializerMeta):
    class Meta:
        types = (list, )

    def serialize(self, list_):
        if not list_:
            return []
        result = []
        for item in list_:
            serializer_cls = PythonSerializer.get_serializer_cls(item)
            serializer = serializer_cls(brief=True, fetch=False)
            result.append(serializer.serialize(item))
        return result

    def serialize_search_result_list(self, list_):
        serializer = SearchSerializer()
        return [serializer.serialize(model) for model in list_]


###################
# model serializers
###################


class SongSerializer(ModelSerializer, SongSerializerMixin,
                     metaclass=SerializerMeta):
    pass


class ArtistSerializer(ModelSerializer, ArtistSerializerMixin,
                       metaclass=SerializerMeta):
    pass


class AlbumSerializer(ModelSerializer, AlbumSerializerMixin,
                      metaclass=SerializerMeta):
    pass


class PlaylistSerializer(ModelSerializer, PlaylistSerializerMixin,
                         metaclass=SerializerMeta):
    pass


class UserSerializer(ModelSerializer, UserSerializerMixin,
                     metaclass=SerializerMeta):
    pass


####################
# object serializers
####################


class SimpleTypeSerializer(PythonSerializer, metaclass=SerializerMeta):
    class Meta:
        types = (str, int, float, type(None))

    def serialize(self, object):
        return object


class ProviderSerializer(PythonSerializer,
                         SimpleSerializerMixin,
                         ProviderSerializerMixin,
                         metaclass=SerializerMeta):
    pass


class SearchSerializer(PythonSerializer, SearchSerializerMixin,
                       metaclass=SerializerMeta):

    def serialize(self, result):
        list_serializer = ListSerializer()
        json_ = {}
        for field, value in self._get_items(result):
            json_[field] = list_serializer.serialize(value)
        return json_



from feeluown.utils import aio
from .app import App


class CliApp(App):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def initialize(self):
        super().initialize()

    def run(self):
        super().run()

        # oncemain should be moved from feeluown/cli to feeluown/app,
        # However, it depends on cli logic temporarily. If there is
        # an common module for these cli logic, oncemain can be moved.
        from feeluown.cli import oncemain  # pylint: disable=cyclic-import

        aio.run_afn(oncemain, self)



from PyQt5.QtWidgets import QAbstractItemView, QFrame, QVBoxLayout

from feeluown.library import SearchType
from feeluown.gui.page_containers.table import TableContainer, Renderer
from feeluown.gui.page_containers.scroll_area import ScrollArea
from feeluown.gui.widgets.img_card_list import ImgCardListDelegate
from feeluown.gui.widgets.songs import SongsTableView, ColumnsMode
from feeluown.gui.base_renderer import TabBarRendererMixin
from feeluown.gui.helpers import BgTransparentMixin
from feeluown.gui.widgets.magicbox import KeySourceIn, KeyType
from feeluown.gui.widgets.header import LargeHeader, MidHeader
from feeluown.gui.widgets.accordion import Accordion
from feeluown.utils.reader import create_reader

Tabs = [('歌曲', SearchType.so),
        ('专辑', SearchType.al),
        ('歌手', SearchType.ar),
        ('歌单', SearchType.pl),
        ('视频', SearchType.vi)]


async def render(req, **kwargs):  # pylint: disable=too-many-locals,too-many-branches
    """/search handler

    :type app: feeluown.app.App
    """
    q = req.query.get('q', '')
    if not q:
        return

    source_in = req.query.get('source_in', None)
    search_type = SearchType(req.query.get('type', SearchType.so.value))
    if source_in is not None:
        source_in = source_in.split(',')
    else:
        source_in = None

    app = req.ctx['app']

    body = Body()
    view = View(app, q)
    body.setWidget(view)
    app.ui.right_panel.set_body(body)

    tab_index = 0
    for i, tab in enumerate(Tabs):
        if tab[1] == search_type:
            tab_index = i
            break

    is_first = True  # Is first search result.
    async for result in app.library.a_search(
            q, type_in=search_type, source_in=source_in):
        if result is not None:
            table_container = TableContainer(app, view.accordion)
            table_container.layout().setContentsMargins(0, 0, 0, 0)

            # HACK: set fixed row for tables.
            # pylint: disable=protected-access
            for table in table_container._tables:
                assert isinstance(table, QAbstractItemView)
                delegate = table.itemDelegate()
                if isinstance(delegate, ImgCardListDelegate):
                    # FIXME: set fixed_row_count in better way.
                    table._fixed_row_count = 2  # type: ignore[attr-defined]
                    delegate.update_settings("card_min_width", 100)
                elif isinstance(table, SongsTableView):
                    table._fixed_row_count = 8
                    table._row_height = table.verticalHeader().defaultSectionSize()

            renderer = SearchResultRenderer(q, tab_index, source_in=source_in)
            await table_container.set_renderer(renderer)
            _, search_type, attrname, show_handler = renderer.tabs[tab_index]
            objects = getattr(result, attrname) or []
            if not objects:  # Result is empty.
                continue

            if search_type is SearchType.so:
                show_handler(  # type: ignore[operator]
                    create_reader(objects), columns_mode=ColumnsMode.playlist)
            else:
                show_handler(create_reader(objects))  # type: ignore[operator]
            source = objects[0].source
            provider = app.library.get(source)
            provider_name = provider.name
            if is_first is False:
                table_container.hide()
            view.accordion.add_section(MidHeader(provider_name), table_container, 6, 12)
            renderer.meta_widget.hide()
            renderer.toolbar.hide()
            is_first = False


class SearchResultRenderer(Renderer, TabBarRendererMixin):
    def __init__(self, q, tab_index, source_in=None):
        self.q = q
        self.tab_index = tab_index
        self.source_in = source_in

        self.tabs = [
            (*Tabs[0], 'songs', self.show_songs),
            (*Tabs[1], 'albums', self.show_albums),
            (*Tabs[2], 'artists', self.show_artists),
            (*Tabs[3], 'playlists', self.show_playlists),
            (*Tabs[4], 'videos', self.show_videos),
        ]

    async def render(self):
        self.render_tab_bar()

    def render_by_tab_index(self, tab_index):
        search_type = self.tabs[tab_index][1]
        self._app.browser.local_storage[KeyType] = search_type.value
        query = {'q': self.q, 'type': search_type.value}
        source_in = self._app.browser.local_storage.get(KeySourceIn, None)
        if source_in is not None:
            query['source_in'] = source_in
        self._app.browser.goto(page='/search', query=query)


class Body(ScrollArea):
    def fillable_bg_height(self):
        """Implement VFillableBg protocol"""
        view = self.widget()
        assert isinstance(view, View)  # make type chckign happy.
        return view.height() - view.accordion.height()


class View(QFrame, BgTransparentMixin):
    def __init__(self, app, q):
        super().__init__()

        self._app = app

        self.title = LargeHeader(f'搜索“{q}”')
        self.accordion = Accordion()

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(20, 0, 20, 0)
        self._layout.setSpacing(0)
        self._layout.addSpacing(30)
        self._layout.addWidget(self.title)
        self._layout.addSpacing(10)
        self._layout.addWidget(self.accordion)
        self._layout.addStretch(0)



from enum import IntEnum, Enum, IntFlag


class ModelFlags(IntFlag):
    none = 0x00000000

    v1 = 0x00000001
    v2 = 0x00000002

    brief = 0x00000010
    normal = brief | 0x00000020


class ModelType(IntEnum):
    dummy = 0

    song = 1
    artist = 2
    album = 3
    playlist = 4
    lyric = 5
    video = 6

    user = 17
    comment = 18

    none = 128


class MediaFlags(IntFlag):
    not_sure = 0b10000000  # 不确定
    not_exists = 0b00  # 不存在相关资源
    sample = 0b01  # 存在试听片段
    free = 0b10  # 存在免费的完整资源
    vip = 0b100  # 存在资源，不免费，vip 可用
    pay = 0b1000  # 存在资源，不免费，付费可用


class SearchType(Enum):
    pl = 'playlist'
    al = 'album'
    ar = 'artist'
    so = 'song'
    vi = 'video'

    @classmethod
    def parse(cls, obj):
        """get member from object

        :param obj: string or SearchType member
        :return: SearchType member

        >>> SearchType.parse('playlist')
        <SearchType.pl: 'playlist'>
        >>> SearchType.parse(SearchType.pl)
        <SearchType.pl: 'playlist'>
        >>> SearchType.parse('xxx')
        Traceback (most recent call last):
          ...
        ValueError: 'xxx' is not a valid SearchType value
        """
        if isinstance(obj, SearchType):
            return obj

        type_aliases_map = {
            cls.pl: ('playlist', 'pl'),
            cls.al: ('album', 'al'),
            cls.ar: ('artist', 'ar'),
            cls.so: ('song', 'so'),
            cls.vi: ('video', 'vi'),
        }
        for type_, aliases in type_aliases_map.items():
            if obj in aliases:
                return type_
        raise ValueError("'%s' is not a valid SearchType value" % obj)

    @classmethod
    def batch_parse(cls, obj):
        """get list of member from obj

        :param obj: obj can be string, list of string or list of member
        :return: list of member

        >>> SearchType.batch_parse('pl,ar')
        [<SearchType.pl: 'playlist'>, <SearchType.ar: 'artist'>]
        >>> SearchType.batch_parse(['pl', 'ar'])
        [<SearchType.pl: 'playlist'>, <SearchType.ar: 'artist'>]
        >>> SearchType.batch_parse('al')
        [<SearchType.al: 'album'>]
        >>> SearchType.batch_parse(SearchType.al)
        [<SearchType.al: 'album'>]
        >>> SearchType.batch_parse([SearchType.al])
        [<SearchType.al: 'album'>]
        """
        if isinstance(obj, SearchType):
            return [obj]
        if isinstance(obj, str):
            return [cls.parse(s) for s in obj.split(',')]
        return [cls.parse(s) for s in obj]


class AlbumType(Enum):
    """Album type enumeration

    中文解释::

        Single 和 EP 会有一些交集，在展示时，会在一起展示，比如 Singles & EPs。
        Compilation 和 Retrospective 也会有交集，展示时，也通常放在一起，统称“合辑”。

    References:

    1. https://www.zhihu.com/question/22888388/answer/33255107
    2. https://zh.wikipedia.org/wiki/%E5%90%88%E8%BC%AF
    """
    standard = 'standard'

    single = 'single'
    ep = 'EP'

    live = 'live'

    compilation = 'compilation'
    retrospective = 'retrospective'

    @classmethod
    def guess_by_name(cls, name):
        """guess album type by its name"""

        # album name which contains following string are `Single`
        #   1. ' - Single'  6+3=9
        #   2. '(single)'   6+2=8
        #   3. '（single）'  6+2=8
        if 'single' in name[-9:].lower():
            return cls.single

        # ' - EP'
        if 'ep' in name[-5:].lower():
            return cls.ep

        if 'live' in name or '演唱会' in name or \
           '音乐会' in name:
            return cls.live

        # '精选集', '精选'
        if '精选' in name[-3:]:
            return cls.retrospective

        return cls.standard



import sys

from PyQt5.QtCore import Qt, QRectF, QRect, QSize
from PyQt5.QtGui import QPalette, QColor, QTextOption, QPainter, \
    QKeySequence, QFont
from PyQt5.QtWidgets import QLabel, QWidget, \
    QVBoxLayout, QSizeGrip, QHBoxLayout, QColorDialog, \
    QMenu, QAction, QFontDialog, QShortcut, QSpacerItem

from feeluown.gui.helpers import esc_hide_widget, resize_font, elided_text
from feeluown.player import LyricLine


IS_MACOS = sys.platform == 'darwin'


def set_bg_color(palette, color):
    palette.setColor(QPalette.Active, QPalette.Window, color)
    palette.setColor(QPalette.Active, QPalette.Base, color)
    palette.setColor(QPalette.Inactive, QPalette.Window, color)
    palette.setColor(QPalette.Inactive, QPalette.Base, color)


def set_fg_color(palette, color):
    palette.setColor(QPalette.Active, QPalette.WindowText, color)
    palette.setColor(QPalette.Active, QPalette.Text, color)
    palette.setColor(QPalette.Inactive, QPalette.WindowText, color)
    palette.setColor(QPalette.Inactive, QPalette.Text, color)


Tooltip = """
* 右键可以弹出设置菜单
* Ctrl+= 或者 Ctrl++ 可以增大字体
* Ctrl+- 可以减小字体
* 鼠标前进后退键可以播放前一首/下一首
* ESC 键可以关闭此歌词窗口
"""


class LyricWindow(QWidget):
    """LyricWindow is a transparent container which contains a real lyric window.

    LyricWindow acts as a transparent container, so the inner window can has
    semi-transparent background. It is also responsible for handling the
    window flags and geometry. It also provides a few APIs for communicating
    with other widgets
    """

    def __init__(self, app):
        super().__init__(parent=None)
        self._app = app

        if IS_MACOS:
            # On macOS, Qt.Tooltip widget can't accept focus and it will hide
            # when the application window is actiavted. Qt.Tool widget can't
            # keep staying on top. Neither of them work well on macOS.
            flags = Qt.WindowStaysOnTopHint | Qt.FramelessWindowHint
        else:
            # TODO: use proper flags on other platforms, see #413 for more details.
            # User can customize the flags in the .fuorc or searchbox, like
            #    app.ui.lyric_windows.setWindowFlags(Qt.xx | Qt.yy)
            flags = Qt.WindowStaysOnTopHint | Qt.FramelessWindowHint | Qt.Tool
        self.setWindowFlags(flags)
        self.setAttribute(Qt.WA_TranslucentBackground)
        self.setToolTip(Tooltip)

        self._inner = InnerLyricWindow(self._app, self)
        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self._inner)

        self._old_pos = None

        esc_hide_widget(self)

    def mousePressEvent(self, e):
        self._old_pos = e.globalPos()

    def mouseMoveEvent(self, e):
        # NOTE: e.button() == Qt.LeftButton don't work on Windows
        # on Windows, even I drag with LeftButton, the e.button() return 0,
        # which means no button
        if self._old_pos is not None:
            delta = e.globalPos() - self._old_pos
            self.move(self.x() + delta.x(), self.y() + delta.y())
            self._old_pos = e.globalPos()

    def mouseReleaseEvent(self, e):
        if not self.rect().contains(e.pos()):
            return
        if e.button() == Qt.BackButton:
            self._app.playlist.previous()
        elif e.button() == Qt.ForwardButton:
            self._app.playlist.next()

    def dump_state(self):
        inner = self._inner
        p = inner.palette()
        geo = self.geometry()
        return {
            'geometry': (geo.x(), geo.y(), geo.width(), geo.height()),
            'font': inner.font().toString(),
            'bg': p.color(QPalette.Active, QPalette.Window).name(QColor.HexArgb),
            'fg': p.color(QPalette.Active, QPalette.Text).name(QColor.HexArgb),
        }

    def apply_state(self, state):
        if not state:
            return

        inner = self._inner

        geo = state.get('geometry')
        if geo:
            self.resize(geo[2], geo[3])
            self.setGeometry(*geo)
        font = inner.font()
        font.fromString(state['font'])
        inner.setFont(font)
        palette = inner.palette()
        set_bg_color(palette, QColor(state['bg']))
        set_fg_color(palette, QColor(state['fg']))
        inner.setPalette(palette)

    def sizeHint(self):
        return QSize(500, 60)

    def resizeEvent(self, e):
        return super().resizeEvent(e)


class SentenceLabel(QLabel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setAlignment(Qt.AlignBaseline | Qt.AlignVCenter | Qt.AlignHCenter)
        self.setWordWrap(False)


class LineLabel(QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.v_spacing = 10
        self.show_trans = True  # Show translated lyric or not.
        self.label = SentenceLabel(parent=self)
        self.trans_label = SentenceLabel(parent=self)
        self.spacer = QSpacerItem(0, 0)

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)

        self._layout.addSpacing(self.v_spacing // 2)
        self._layout.addWidget(self.label)
        self._layout.addSpacerItem(self.spacer)
        self._layout.addWidget(self.trans_label)
        self._layout.addSpacing(self.v_spacing // 2)

        line = LyricLine('...', '...', False)
        self.set_line(line)
        # The default size(calculated by Qt) may be different from the size
        # calculated by line_sizehint. Remember to specify size at very first,
        # otherwise it may show with the default size(calculated by Qt).
        self.resize(self.line_sizehint(line))

    def toggle_show_trans(self):
        self.show_trans = not self.show_trans
        if self.show_trans:
            self.trans_label.show()
            self.spacer.changeSize(0, self.v_spacing//3)
        else:
            self.trans_label.hide()
            self.spacer.changeSize(0, 0)

    def set_line(self, line: LyricLine):
        self.label.setText(
            elided_text(line.origin, self.width(), self.font()))
        if self.show_trans and line.has_trans:
            self.trans_label.show()
            self.trans_label.setText(
                elided_text(line.trans, self.width(), self.font()))
        else:
            self.trans_label.hide()

    def line_sizehint(self, line: LyricLine):
        """Proper size to show the line."""
        rect = self.label.fontMetrics().boundingRect(line.origin)
        height = rect.height()
        if self.show_trans and line.has_trans:
            height = height * 2
        # Sometimes width is not enough for text, so add buffer.
        h_buffer = rect.height()
        # Add some padding for vertical so that it looks more beautiful.
        v_buffer = rect.height() // 4
        height += self.v_spacing + v_buffer + self.spacer.geometry().height()
        return QSize(rect.width() + h_buffer, height)

    def setFont(self, font):
        super().setFont(font)

        self.label.setFont(font)
        font2 = QFont(font)
        font2.setBold(False)
        if font.pointSize() != 0:
            resize_font(font2, -4)
        else:
            resize_font(font2, -8)
        self.trans_label.setFont(font2)

    def setPalette(self, palette):
        super().setPalette(palette)
        self.label.setPalette(palette)
        self.trans_label.setPalette(palette)


class InnerLyricWindow(QWidget):
    """
    This window is responsible for rendering one line of a lyric.
    This window need not to know which song is playing, or if
    the song is changed.
    """

    def __init__(self, app, parent=None):
        super().__init__(parent=parent)
        self._app = app

        self._border_radius = 0
        # When _auto_resize is True,
        # the window size adapts to the lyric sentence width.
        self._auto_resize = True
        self._size_grip = QSizeGrip(self)
        self.line_label = LineLabel(self)

        self._app.live_lyric.line_changed.connect(self.set_line)
        self._app.live_lyric.lyrics_changed.connect(self.on_lyrics_changed)
        QShortcut(QKeySequence.ZoomIn, self).activated.connect(self.zoomin)
        QShortcut(QKeySequence.ZoomOut, self).activated.connect(self.zoomout)
        QShortcut(QKeySequence('Ctrl+='), self).activated.connect(self.zoomin)

        self._layout = QHBoxLayout(self)
        self.setup_ui()

    def setup_ui(self):
        if self._auto_resize:
            self._size_grip.hide()
        self.on_font_size_changed()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self.line_label)
        self._layout.addWidget(self._size_grip)
        self._layout.setAlignment(self._size_grip, Qt.AlignBottom)

    def set_line(self, line: LyricLine):
        # Ignore updating when the window is invisible.
        if not self.isVisible():
            return
        size = self.line_label.line_sizehint(line)
        self.line_label.resize(size)
        if self._auto_resize:
            self_size = QSize(size.width(), size.height())
            self.resize(self_size)
            self.parent().resize(self_size)  # type: ignore
            self.parent().updateGeometry()   # type: ignore
        self.line_label.set_line(line)

    def on_lyrics_changed(self, lyric, *_):
        if lyric is None:
            self.set_line(LyricLine('未找到可用歌词', '', False))

    def zoomin(self):
        label = self.line_label
        font = label.font()
        resize_font(font, +1)
        label.setFont(font)

    def zoomout(self):
        label = self.line_label
        font = label.font()
        resize_font(font, - 1)
        label.setFont(font)

    def on_font_size_changed(self):
        self._border_radius = self.fontMetrics().height() // 3
        width = max(1, self._border_radius * 2)
        self._size_grip.setFixedWidth(width)

    def paintEvent(self, e):
        """Draw shapes to make the size_grip more obvious.

        Note the shapes can't be drawed on the outside container (LyricWindow)
        due to it sets the attribute WA_TranslucentBackground.
        """
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing)
        painter.setPen(Qt.NoPen)
        painter.setBrush(self.palette().color(QPalette.Window))
        painter.drawRoundedRect(self.rect(), self._border_radius, self._border_radius)

        # Draw an circle button to indicate that the window can be resized.
        if self._auto_resize:
            return
        painter.save()
        painter.setPen(QColor('white'))
        option = QTextOption()
        option.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        rect = QRect(self.mapToParent(self._size_grip.pos()), self._size_grip.size())
        painter.drawText(QRectF(rect), '●', option)
        painter.restore()

    def setPalette(self, a0: QPalette) -> None:
        super().setPalette(a0)
        self.line_label.setPalette(a0)

    def setFont(self, a0: QFont) -> None:
        super().setFont(a0)
        self.line_label.setFont(a0)
        self.on_font_size_changed()

    def show_color_dialog(self, bg=True):
        def set_color(color):
            palette = self.palette()
            if bg:
                set_bg_color(palette, color)
            else:
                set_fg_color(palette, color)
            # Note that this widget(self) must also set the palette,
            # so the background can work as expected.
            self.setPalette(palette)

        dialog = QColorDialog(self)
        # Set WA_DeleteOnClose so that the dialog can be deleted (from self.children).
        dialog.setAttribute(Qt.WA_DeleteOnClose)
        if bg:
            color = self.palette().color(QPalette.Window)
        else:
            color = self.palette().color(QPalette.Text)
        dialog.setCurrentColor(color)
        dialog.currentColorChanged.connect(set_color)
        dialog.colorSelected.connect(set_color)
        dialog.setOption(QColorDialog.ShowAlphaChannel, True)
        # On KDE(with Xorg), if the dialog is in modal state,
        # the window is dimming.
        if sys.platform == 'linux':
            dialog.show()
        else:
            dialog.open()

    def show_font_dialog(self):
        dialog = QFontDialog(self.font(), self)
        # Set WA_DeleteOnClose so that the dialog can be deleted (from self.children).
        dialog.setAttribute(Qt.WA_DeleteOnClose)
        dialog.currentFontChanged.connect(self.setFont)
        dialog.fontSelected.connect(self.setFont)
        dialog.open()

    def toggle_auto_resize(self):
        self._auto_resize = not self._auto_resize
        if self._auto_resize:
            self._size_grip.hide()
        else:
            self._size_grip.show()

    def contextMenuEvent(self, e):
        menu = QMenu()
        bg_color_action = QAction('背景颜色', menu)
        fg_color_action = QAction('文字颜色', menu)
        font_action = QAction('字体', menu)
        toggle_trans_action = QAction('双语歌词', menu)
        toggle_trans_action.setCheckable(True)
        toggle_trans_action.setChecked(self.line_label.show_trans)
        toggle_fiexed_size_action = QAction('大小自动', menu)
        toggle_fiexed_size_action.setCheckable(True)
        toggle_fiexed_size_action.setChecked(self._auto_resize)
        menu.addAction(bg_color_action)
        menu.addAction(fg_color_action)
        menu.addSeparator()
        menu.addAction(font_action)
        menu.addSeparator()
        menu.addAction(toggle_trans_action)
        menu.addAction(toggle_fiexed_size_action)

        bg_color_action.triggered.connect(lambda: self.show_color_dialog(bg=True))
        fg_color_action.triggered.connect(lambda: self.show_color_dialog(bg=False))
        font_action.triggered.connect(self.show_font_dialog)
        toggle_trans_action.triggered.connect(self.line_label.toggle_show_trans)
        toggle_fiexed_size_action.triggered.connect(self.toggle_auto_resize)

        menu.exec(e.globalPos())

    def showEvent(self, e) -> None:
        self.set_line(self._app.live_lyric.current_line)
        return super().showEvent(e)



import logging
import json
from contextlib import contextmanager
from typing import Optional, Type

from feeluown.consts import STATE_FILE
from feeluown.utils.request import Request
from feeluown.library import Library
from feeluown.utils.dispatch import Signal
from feeluown.library import (
    Resolver, reverse, resolve,
    ResolverNotFound, ResolveFailed,
)
from feeluown.player import (
    PlaybackMode, Playlist, LiveLyric,
    FM, Player, RecentlyPlayed, PlayerPositionDelegate
)
from feeluown.plugin import plugins_mgr
from feeluown.version import VersionManager
from feeluown.task import TaskManager

from .mode import AppMode


logger = logging.getLogger(__name__)


class App:
    """App 基类"""
    _instance = None

    # .. deprecated:: 3.8
    #    Use :class:`AppMode` instead.
    DaemonMode = AppMode.server.value
    GuiMode = AppMode.gui.value
    CliMode = AppMode.cli.value

    def __init__(self, args, config, **kwargs):
        App._instance = self
        self.mode = config.MODE  # DEPRECATED: use app.config.MODE instead
        self.config = config
        self.args = args
        self.plugin_mgr = plugins_mgr

        self.initialized = Signal()
        # .. versionadded:: 3.8.11
        #    App started signal. Maybe the initialized signal can be removed?
        self.started = Signal()  # App is ready to use, for example, UI is available.
        self.about_to_shutdown = Signal()

        self.request = Request()  # TODO: rename request to http
        self.version_mgr = VersionManager(self)
        self.task_mgr = TaskManager(self)

        # Library.
        self.library = Library(config.PROVIDERS_STANDBY)
        # TODO: initialization should be moved into initialize
        Resolver.library = self.library
        # Player.
        self.player = Player(
            audio_device=bytes(config.MPV_AUDIO_DEVICE, 'utf-8'),
            fade=config.PLAYBACK_CROSSFADE,
            fade_time_ms=config.PLAYBACK_CROSSFADE_DURATION,
        )
        # Theoretically, each caller maintain its own position delegate.
        # For simplicity, this delegate is created for common use cases and
        #
        # For progress slider and label in player bar, user may only see the difference
        # when it is greater than 1s. There are usually only one or two lines of lyrics
        # in one second. Considering the use cases and performance, I guess 300ms is
        # a reasonable interval.
        self.player_pos_per300ms = PlayerPositionDelegate(self.player, interval=300)
        self.playlist = Playlist(self, audio_select_policy=config.AUDIO_SELECT_POLICY)
        self.live_lyric = LiveLyric(self)
        self.fm = FM(self)
        self.recently_played = RecentlyPlayed(self.playlist)

        # TODO: initialization should be moved into initialize
        self.player.set_playlist(self.playlist)

        self.about_to_shutdown.connect(lambda _: self.dump_and_save_state(), weak=False)

    def initialize(self):
        self.player_pos_per300ms.initialize()
        self.player_pos_per300ms.changed.connect(self.live_lyric.on_position_changed)
        self.playlist.song_changed.connect(self.live_lyric.on_song_changed,
                                           aioqueue=True)
        self.player.media_loading_failed.connect(
            lambda *args: self.show_msg('播放器加载资源失败'), weak=False, aioqueue=True)
        self.plugin_mgr.enable_plugins(self)

    def run(self):
        pass

    @property
    def has_server(self) -> bool:
        """
        .. versionadded:: 3.8
        """
        return AppMode.server in AppMode(self.config.MODE)

    @property
    def has_gui(self) -> bool:
        """
        .. versionadded:: 3.8
        """
        return AppMode.gui in AppMode(self.config.MODE)

    def show_msg(self, msg, *args, **kwargs):
        """在程序中显示消息，一般是用来显示程序当前状态"""
        # pylint: disable=no-self-use, unused-argument
        logger.info(msg)

    def get_listen_addr(self):
        return '0.0.0.0' if self.config.ALLOW_LAN_CONNECT else '127.0.0.1'

    def apply_state(self, state):
        playlist = self.playlist
        player = self.player
        recently_played = self.recently_played

        if state:
            player.volume = state['volume']

            # Restore recently_played states.
            recently_played_models = []
            for model in state.get('recently_played', []):
                try:
                    model = resolve(model)
                except ResolverNotFound:
                    pass
                except ResolveFailed as e:
                    logger.warning(f'resolve failed, {e}')
                else:
                    recently_played_models.append(model)
            recently_played.init_from_models(recently_played_models)

            # Restore playlist states.
            playlist.playback_mode = PlaybackMode(state['playback_mode'])
            songs = []
            for song in state['playlist']:
                try:
                    song = resolve(song)
                except ResolverNotFound:
                    pass
                except ResolveFailed as e:
                    logger.warning(f'resolve failed, {e}')
                else:
                    songs.append(song)
            playlist.set_models(songs)
            song = state['song']

            def before_media_change(old_media, media):
                # When the song has no media or preparing media is failed,
                # the current_song is not the song we set.
                #
                # When user play a new media directly through player.play interface,
                # the old media is not None.
                if old_media is not None or playlist.current_song != song:
                    player.media_about_to_changed.disconnect(before_media_change)
                    player.set_play_range()

            if song is not None:
                try:
                    song = resolve(state['song'])
                except ResolverNotFound:
                    pass
                else:
                    player.media_about_to_changed.connect(before_media_change,
                                                          weak=False)
                    player.pause()
                    player.set_play_range(start=state['position'])
                    playlist.set_current_song(song)

    def load_state(self):
        try:
            with open(STATE_FILE, 'r', encoding='utf-8') as f:
                state = json.load(f)
        except FileNotFoundError:
            return {}
        except json.decoder.JSONDecodeError:
            logger.exception('invalid state file')
            return {}
        return state

    def load_and_apply_state(self):
        state = self.load_state()
        self.apply_state(state)

    def dump_state(self):
        playlist = self.playlist
        player = self.player
        recently_played = self.recently_played

        song = self.player.current_song
        if song is not None:
            song = reverse(song, as_line=True)
        # TODO: dump player.media
        state = {
            'playback_mode': playlist.playback_mode.value,
            'volume': player.volume,
            'state': player.state.value,
            'song': song,
            # cast position to int to avoid such value 2.7755575615628914e-17
            'position': int(player.position or 0),
            'playlist': [reverse(song, as_line=True) for song in playlist.list()],
            'recently_played': [reverse(song, as_line=True)
                                for song in recently_played.list_songs()]
        }
        return state

    def dump_and_save_state(self):
        logger.info("Dump and save app state")
        state = self.dump_state()
        with open(STATE_FILE, 'w', encoding='utf-8') as f:
            json.dump(state, f)

    @contextmanager
    def create_action(self, s):  # pylint: disable=no-self-use
        """根据操作描述生成 Action (alpha)

        设计缘由：用户需要知道目前程序正在进行什么操作，进度怎么样，
        结果是失败或者成功。这里将操作封装成 Action。
        """
        show_msg = self.show_msg

        class ActionError(Exception):
            pass

        class Action:
            def set_progress(self, value):
                value = int(value * 100)
                show_msg(s + f'...{value}%', timeout=5000)

            def failed(self, msg=''):
                raise ActionError(msg)

        show_msg(s + '...', timeout=5000)  # doing
        try:
            yield Action()
        except ActionError as e:
            show_msg(s + f'...failed\t{str(e)}')
        except Exception as e:
            show_msg(s + f'...error\t{str(e)}')  # error
            raise
        else:
            show_msg(s + '...done')  # done

    def about_to_exit(self):
        logger.info('Do graceful shutdown')
        try:
            self.about_to_shutdown.emit(self)
            self.player_pos_per300ms.stop()
            self.player.stop()
            self.exit_player()
        except:  # noqa, pylint: disable=bare-except
            logger.exception("about-to-exit failed")
        logger.info('Ready for shutdown, or crash :)')

    def exit_player(self):
        self.player.shutdown()  # this cause 'abort trap' on macOS

    def exit(self):
        self.about_to_exit()


def get_app() -> Optional['App']:
    """
    .. versionadded:: 3.8.11
    """
    return App._instance


def create_app(args, config) -> App:
    """App factory function.

    Do not add `create` function to :class:`App` because QWidget also has
    a `create` function.
    """
    need_server = AppMode.server in AppMode(config.MODE)
    need_window = AppMode.gui in AppMode(config.MODE)

    # pylint: disable=import-outside-toplevel,cyclic-import
    cls: Type[App]

    if args.cmd is not None:
        from feeluown.app.cli_app import CliApp
        cls = CliApp
    elif need_server and need_window:
        from feeluown.app.mixed_app import MixedApp
        cls = MixedApp
    elif need_window:
        from feeluown.app.gui_app import GuiApp
        cls = GuiApp
    elif need_server:
        from feeluown.app.server_app import ServerApp
        cls = ServerApp
    else:
        cls = App
    return cls(args, config)



import asyncio
import logging
import warnings
from abc import ABCMeta, abstractmethod
from enum import IntEnum

from feeluown.utils.dispatch import Signal
# some may import `Playlist` and `PlaybackMode` from player module
from feeluown.player.playlist import PlaybackMode, Playlist
from .metadata import Metadata

__all__ = ('Playlist',
           'AbstractPlayer',
           'PlaybackMode',
           'State',)

logger = logging.getLogger(__name__)


class State(IntEnum):
    """
    Player states.
    """
    stopped = 0
    #: 处于 paused 状态时，current_song 也可能为 False
    paused = 1
    playing = 2


class AbstractPlayer(metaclass=ABCMeta):
    """Player abstrace base class.

    Note that signals may be emitted from different thread. You should
    take care of race condition.
    """

    def __init__(self, _=None, **kwargs):
        """
        :param _: keep this arg to keep backward compatibility
        """
        self._position = 0  # seconds
        self._volume = 100  # (0, 100)
        self._playlist = None
        self._state = State.stopped
        self._duration = None

        self._current_media = None
        self._current_metadata = Metadata()
        self._video_format = None

        #: player position changed signal
        self.position_changed = Signal()
        self.seeked = Signal()

        #: player state changed signal
        self.state_changed = Signal()

        #: duration changed signal
        self.duration_changed = Signal()

        #: media about to change: (old_media, media)
        self.media_about_to_changed = Signal()
        self.media_changed = Signal()  # Media source is changed (not loaded yet).
        # The difference between media_loaded and media_loaded_v2 is that
        # media_loaded_v2 carries some media properties.
        self.media_loaded = Signal()  # Start to play the media.
        self.media_loaded_v2 = Signal()  # emit(properties)
        # Metadata is changed, and it may be changed during playing.
        self.metadata_changed = Signal()
        self.media_finished = Signal()  # Finish to play the media.
        self.media_loading_failed = Signal()

        #: volume changed signal: (int)
        self.volume_changed = Signal()

    @property
    def state(self):
        """Player state

        :rtype: State
        """
        return self._state

    @state.setter
    def state(self, value):
        """set player state, emit state changed signal

        outer object should not set state directly,
        use ``pause`` / ``resume`` / ``stop`` / ``play`` method instead.
        """
        self._state = value
        self.state_changed.emit(value)

    @property
    def current_media(self):
        return self._current_media

    @property
    def current_metadata(self) -> Metadata:
        """Metadata for the current media

        Check `MetadataFields` for all possible fields. Note that some fields
        can be missed if them are unknown. For example, a video's metadata
        may have no genre info.
        """
        return self._current_metadata

    @property
    def position(self):
        """player position, the units is seconds"""
        return self._position

    @position.setter
    def position(self, position):
        """set player position, the units is seconds"""

    @property
    def volume(self):
        return self._volume

    @volume.setter
    def volume(self, value):
        value = 0 if value < 0 else value
        value = 100 if value > 100 else value
        self._volume = value
        self.volume_changed.emit(value)

    @property
    def duration(self):
        """player media duration, the units is seconds"""
        return self._duration

    @duration.setter
    def duration(self, value):
        value = value or 0
        if value != self._duration:
            self._duration = value
            self.duration_changed.emit(value)

    @abstractmethod
    def play(self, media, video=True, metadata=None):
        """play media

        :param media: a local file absolute path, or a http url that refers to a
            media file
        :param video: show video or not
        :param metadata: metadata for the media
        """

    @abstractmethod
    def set_play_range(self, start=None, end=None):
        pass

    @abstractmethod
    def resume(self):
        """play playback"""

    @abstractmethod
    def pause(self):
        """pause player"""

    @abstractmethod
    def toggle(self):
        """toggle player state"""

    @abstractmethod
    def stop(self):
        """stop player"""

    @abstractmethod
    def shutdown(self):
        """shutdown player, do some clean up here"""

    # ------------------
    # Deprecated methods
    # ------------------
    @property
    def playlist(self):
        """(DEPRECATED) player playlist

        Player SHOULD not know the existence of playlist. However, in the
        very beginning, the player depends on playlist and listen playlist's
        signal. Other programs may depends on the playlist property and
        we keep it for backward compatibility.

        TODO: maybe add a DeprecationWarning in v3.8.

        :return: :class:`.Playlist`
        """
        return self._playlist

    def set_playlist(self, playlist):
        self._playlist = playlist

    @property
    def current_song(self):
        """(Deprecated) alias of playlist.current_song

        Please use playlist.current_song instead.
        """
        warnings.warn('use playlist.current_model instead', DeprecationWarning)
        return self._playlist.current_song

    def load_song(self, song) -> asyncio.Task:
        """加载歌曲

        如果目标歌曲与当前歌曲不相同，则修改播放列表当前歌曲，
        播放列表会发出 song_changed 信号，player 监听到信号后调用 play 方法，
        到那时才会真正的播放新的歌曲。如果和当前播放歌曲相同，则忽略。

        .. note::

            调用方应该直接调用 playlist.current_song = song 来切换歌曲
        """
        assert song is not None
        warnings.warn(
            'use playlist.set_current_model instead, this will be removed in v3.8',
            DeprecationWarning
        )
        return self._playlist.set_current_song(song)

    def play_song(self, song):
        """加载并播放指定歌曲"""
        warnings.warn(
            'use playlist.set_current_model instead, this will be removed in v3.8',
            DeprecationWarning
        )
        return self._playlist.set_current_song(song)

    def play_songs(self, songs):
        """(alpha) play list of songs"""
        warnings.warn(
            'use playlist.init_from instead, this will be removed in v3.8',
            DeprecationWarning
        )
        self.playlist.set_models(songs, next_=True)



import base64
import itertools
import logging
import os
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Dict, Iterable, List

import tomlkit

from feeluown.consts import COLLECTIONS_DIR
from feeluown.utils.dispatch import Signal
from feeluown.library import resolve, reverse, ResolverNotFound, \
    ResolveFailed, ModelState
from feeluown.utils.utils import elfhash

logger = logging.getLogger(__name__)

COLL_LIBRARY_IDENTIFIER = 'library'
COLL_POOL_IDENTIFIER = 'pool'
# for backward compat, we should never change these filenames
LIBRARY_FILENAME = f'{COLL_LIBRARY_IDENTIFIER}.fuo'
POOL_FILENAME = f'{COLL_POOL_IDENTIFIER}.fuo'
DEPRECATED_FUO_FILENAMES = (
    'Songs.fuo', 'Albums.fuo', 'Artists.fuo', 'Videos.fuo'
)

TOML_DELIMLF = "+++\n"


class CollectionAlreadyExists(Exception):
    pass


class CollectionType(Enum):
    sys_library = 16
    sys_pool = 13

    mixed = 8


class Collection:

    def __init__(self, fpath):
        # TODO: 以后考虑添加 identifier 字段，identifier
        # 字段应该尽量设计成可以跨电脑使用
        self.fpath = str(fpath)
        # TODO: 目前还没想好 collection identifier 计算方法，故添加这个函数
        # 现在把 fpath 当作 identifier 使用，但对外透明
        self.identifier = elfhash(base64.b64encode(bytes(self.fpath, 'utf-8')))

        # these variables should be inited during loading
        self.type = None
        self.name = None  # Collection title.
        self.models = []
        self.updated_at = None
        self.created_at = None
        self.description = None
        self._has_nonexistent_models = False

        #: tomkit.toml_document.Document
        self._metadata = None

    def load(self):
        """解析文件，初始化自己"""
        # pylint: disable=too-many-branches
        self.models = []
        filepath = Path(self.fpath)
        name = filepath.stem
        stat_result = filepath.stat()
        self.updated_at = datetime.fromtimestamp(stat_result.st_mtime)
        self.name = name
        if name == COLL_LIBRARY_IDENTIFIER:
            self.type = CollectionType.sys_library
        elif name == COLL_POOL_IDENTIFIER:
            self.type = CollectionType.sys_pool
        else:
            self.type = CollectionType.mixed

        # parse file content
        with filepath.open(encoding='utf-8') as f:
            first = f.readline()
            lines = []
            if first == TOML_DELIMLF:
                is_valid = False
                for line in f:
                    if line == TOML_DELIMLF:
                        is_valid = True
                        break
                    lines.append(line)
                if is_valid is True:
                    toml_str = ''.join(lines)
                    metadata = tomlkit.parse(toml_str)
                    self._loads_metadata(metadata)
                    lines = []
                else:
                    logger.warning('the metadata is invalid, will ignore it')
            else:
                lines.append(first)

            for line in itertools.chain(lines, f):
                if not line.strip():  # ignore empty lines
                    continue
                try:
                    model = resolve(line)
                except ResolverNotFound:
                    logger.warning('resolver not found for line:%s', line)
                    model = None
                except ResolveFailed as e:
                    logger.warning('resolve failed, file:%s, line:%s, error:%s',
                                   str(filepath), line, str(e))
                    model = None
                if model is not None:
                    if model.state is ModelState.not_exists:
                        self._has_nonexistent_models = True
                    self.models.append(model)

    @classmethod
    def create_empty(cls, fpath, title=''):
        """Create an empty collection."""
        if os.path.exists(fpath):
            raise CollectionAlreadyExists()

        doc = tomlkit.document()
        if title:
            doc.add('title', title)         # type: ignore[arg-type]
        doc.add('created', datetime.now())  # type: ignore[arg-type]
        doc.add('updated', datetime.now())  # type: ignore[arg-type]
        with open(fpath, 'w', encoding='utf-8') as f:
            f.write(TOML_DELIMLF)
            f.write(tomlkit.dumps(doc))
            f.write(TOML_DELIMLF)

        coll = cls(fpath)
        coll._loads_metadata(doc)
        coll.type = CollectionType.mixed
        return coll

    def add(self, model):
        """add model to collection

        :param model: :class:`feeluown.library.BaseModel`
        :return: True means succeed, False means failed
        """
        if model not in self.models:
            line = reverse(model, as_line=True)
            with open(self.fpath, 'r+', encoding='utf-8') as f:
                content = f.read()
                parts = content.split(TOML_DELIMLF, maxsplit=2)
                body = parts[-1]
                f.seek(0, 0)
                self._write_metadata_if_needed(f)
                f.write(f'{line}\n{body}')
                f.truncate()
            self.models.insert(0, model)
        return True

    def remove(self, model):
        if model in self.models:
            url = reverse(model)
            with open(self.fpath, 'r+', encoding='utf-8') as f:
                content = f.read()
                parts = content.split(TOML_DELIMLF, maxsplit=2)
                body = parts[-1]
                lines = []
                for line in body.split('\n'):
                    if line.startswith(url):
                        continue
                    if line:
                        lines.append(line)

                f.seek(0)
                self._write_metadata_if_needed(f)
                f.write('\n'.join(lines))
                f.truncate()
                # 确保最后写入一个换行符，让文件更加美观
                if lines and not lines[-1].endswith('\n'):
                    f.write('\n')
            self.models.remove(model)
        return True

    def on_provider_added(self, provider):
        if not self._has_nonexistent_models:
            return
        for i, model in enumerate(self.models.copy()):
            if model.state is ModelState.not_exists and \
                    model.source == provider.identifier:
                new_model = resolve(reverse(model, as_line=True))
                # TODO: emit data changed signal
                self.models[i] = new_model
        # TODO: set _has_nonexistent_models to proper value

    def on_provider_removed(self, provider):
        for model in self.models:
            if model.source == provider.identifier:
                model.state = ModelState.not_exists
                self._has_nonexistent_models = True

    def _loads_metadata(self, metadata):
        self._metadata = metadata
        self.created_at = metadata.get('created')
        self.updated_at = metadata.get('updated', self.updated_at)
        self.name = metadata.get('title', self.name)
        self.description = metadata.get('description')

    def _write_metadata_if_needed(self, f):
        # write metadata only if it had before
        if self._metadata:
            self.updated_at = self._metadata['updated'] = datetime.now()
            s = tomlkit.dumps(self._metadata)
            f.write(TOML_DELIMLF)
            f.write(s)
            if not s.endswith('\n'):
                f.write('\n')
            f.write(TOML_DELIMLF)


class CollectionManager:

    def __init__(self, app):
        self._app = app
        self.scan_finished = Signal()
        self._library = app.library
        self.default_dir = COLLECTIONS_DIR

        self._id_coll_mapping: Dict[int, Collection] = {}
        self._sys_colls = {}

    def get(self, identifier):
        if identifier in (CollectionType.sys_pool, CollectionType.sys_library):
            return self._sys_colls[identifier]
        return self._id_coll_mapping[int(identifier)]

    def get_coll_library(self):
        for coll in self._id_coll_mapping.values():
            if coll.type == CollectionType.sys_library:
                return coll
        assert False, "collection 'library' must exists."

    def create(self, fname, title) -> Collection:
        first_valid_dir = ''
        for d, exists in self._get_dirs():
            if exists:
                first_valid_dir = d
                break

        assert first_valid_dir, 'there must be a valid collection dir'
        normalized_name = fname.replace(' ', '_')
        fpath = os.path.join(first_valid_dir, normalized_name)
        filepath = f'{fpath}.fuo'
        logger.info(f'Create collection:{title} at {filepath}')
        return Collection.create_empty(filepath, title)

    def remove(self, collection: Collection):
        coll_id = collection.identifier
        if coll_id in self._id_coll_mapping:
            self._id_coll_mapping.pop(coll_id)
            os.remove(collection.fpath)

    def _get_dirs(self, ):
        directorys = [self.default_dir]
        if self._app.config.COLLECTIONS_DIR:
            if isinstance(self._app.config.COLLECTIONS_DIR, list):
                directorys += self._app.config.COLLECTIONS_DIR
            else:
                directorys.append(self._app.config.COLLECTIONS_DIR)
        expanded_dirs = []
        for directory in directorys:
            directory = os.path.expanduser(directory)
            expanded_dirs.append((directory, os.path.exists(directory)))
        return expanded_dirs

    def scan(self):
        colls: List[Collection] = []
        for coll in self._scan():
            if coll.type == CollectionType.sys_library:
                self._sys_colls[CollectionType.sys_library] = coll
            elif coll.type == CollectionType.sys_pool:
                self._sys_colls[CollectionType.sys_pool] = coll
            else:
                colls.append(coll)

        if CollectionType.sys_pool not in self._sys_colls:
            pool_fpath = os.path.join(self.default_dir, POOL_FILENAME)
            assert not os.path.exists(pool_fpath)
            logger.info('Generating collection pool.')
            coll = Collection.create_empty(pool_fpath, '想听')
            self._sys_colls[CollectionType.sys_pool] = coll

        pool_coll = self._sys_colls[CollectionType.sys_pool]
        library_coll = self._sys_colls[CollectionType.sys_library]
        colls.insert(0, pool_coll)
        colls.insert(0, library_coll)
        for collection in colls:
            coll_id = collection.identifier
            assert coll_id not in self._id_coll_mapping, collection.fpath
            self._id_coll_mapping[coll_id] = collection
        self.scan_finished.emit()

    def refresh(self):
        self.clear()
        self.scan()

    def listall(self):
        return self._id_coll_mapping.values()

    def clear(self):
        self._id_coll_mapping.clear()

    def _scan(self) -> Iterable[Collection]:
        """Scan collections directories for valid fuo files, yield
        Collection instance for each file.
        """
        default_fpaths = []
        valid_dirs = self._get_dirs()
        for directory, exists in valid_dirs:
            if not exists:
                logger.warning('Collection directory %s does not exist', directory)
                continue
            for filename in os.listdir(directory):
                if not filename.endswith('.fuo'):
                    continue
                filepath = os.path.join(directory, filename)
                if filename in DEPRECATED_FUO_FILENAMES:
                    default_fpaths.append(filepath)
                    continue
                coll = Collection(filepath)
                coll.load()
                self._app.library.provider_added.connect(coll.on_provider_added)
                self._app.library.provider_removed.connect(coll.on_provider_removed)
                yield coll

        fpath, generated = self.generate_library_coll_if_needed(default_fpaths)
        # Avoid to yield a duplicated collection.
        if generated is True:
            coll = Collection(fpath)
            coll.load()
            self._app.library.provider_added.connect(coll.on_provider_added)
            self._app.library.provider_removed.connect(coll.on_provider_removed)
            yield coll

    def generate_library_coll_if_needed(self, default_fpaths):
        """
        :return: (library file path, generated)
        """
        library_fpath = os.path.join(self.default_dir, LIBRARY_FILENAME)
        if os.path.exists(library_fpath):
            if default_fpaths:
                paths_str = ','.join(default_fpaths)
                logger.warning(f'{paths_str} are ignored since {library_fpath} exists')
            return library_fpath, False

        logger.info('Generating collection library')
        lines = [TOML_DELIMLF,
                 'title = "Library"',
                 TOML_DELIMLF]
        if default_fpaths:
            for fpath in default_fpaths:
                with open(fpath, 'r', encoding='utf-8') as f:
                    for line in f:
                        if line.startswith('fuo://'):
                            lines.append(line)
        with open(library_fpath, 'w', encoding='utf-8') as f:
            f.write('\n'.join(lines))

        # backup old sys collections if they exists
        for fpath in default_fpaths:
            dirname = os.path.dirname(fpath)
            filename = os.path.basename(fpath)
            filename += '.bak'
            new_fpath = os.path.join(dirname, filename)
            logger.info(f'Rename {fpath} to {new_fpath}')
            os.rename(fpath, new_fpath)
        return library_fpath, True



from collections.abc import MutableMapping
from enum import Enum


class MetadataFields(Enum):
    """
    Check the following docs for fields definition
    0. id3tag
    1. www.freedesktop.org -> mpris-spec/metadata/#fields
    2. developer.apple.com -> MPMediaItem
    """
    # setby refers to how the metadata is generated/gotten.
    setby = '__setby__'

    uri = 'uri'  # uri with `fuo` scheme

    title = 'title'
    artists = 'artists'  # The value is a list of strings.
    album = 'album'
    year = 'year'
    genre = 'genre'
    track = 'track'  # The track number on the album disc.
    source = 'source'
    artwork = 'artwork'  # The album/video cover iamge url.
    released = 'released'  # The released date of the song or it's album.


class Metadata(MutableMapping):
    """Metadata is a dict that transform the key to MetadataFields.

    >>> m = Metadata({'title': 'hello world'})
    >>> m['title']
    'hello world'
    >>> m.get('notexist', 'notexist')
    'notexist'
    """
    def __init__(self, *args, **kwargs):
        self._store = dict()
        self.update(dict(*args, **kwargs))

    def __getitem__(self, name):
        return self._store[self._to_field(name)]

    def __contains__(self, name):
        return self._to_field(name) in self._store

    def __setitem__(self, key, value):
        field = self._to_field(key)

        # Validate the type of a value
        field_type_mapping = {
            MetadataFields.artists: list,
        }
        expected = field_type_mapping.get(field, str)
        if not isinstance(value, expected):
            raise ValueError(f'field {field} expect {expected}, acture {type(value)}')
        self._store[field] = value

    def __delitem__(self, key):
        del self._store[self._to_field(key)]

    def __iter__(self):
        return iter(self._store)

    def __len__(self):
        return len(self._store)

    def _to_field(self, name):
        try:
            field = MetadataFields(name)
        except ValueError:
            raise KeyError(f'invalid key {name}') from None
        else:
            return field



"""
all metadata related widgets, for example: cover, and so on.
"""

from PyQt5.QtCore import Qt, QSize
from PyQt5.QtGui import QFont
from PyQt5.QtWidgets import QLabel, QHBoxLayout, QVBoxLayout, \
    QSpacerItem, QFrame, QSizePolicy

from feeluown.gui.helpers import elided_text
from .cover_label import CoverLabelV2


class getset_property:
    def __init__(self, name):
        self.name = name
        self.name_real = '_' + name

    def __get__(self, instance, owner):
        if hasattr(instance, self.name_real):
            return getattr(instance, self.name_real)
        return None

    def __set__(self, instance, value):
        setattr(instance, self.name_real, value)
        instance.on_property_updated(self.name)


class MetaWidget(QFrame):

    def clear(self):
        self.title = None
        self.subtitle = None
        self.source = None
        self.cover = None
        self.created_at = None
        self.updated_at = None
        self.creator = None
        self.songs_count = None
        self.released_at = None

    def on_property_updated(self, name):
        pass

    # TODO: use metaclass
    title = getset_property('title')
    subtitle = getset_property('subtitle')
    source = getset_property('source')
    cover = getset_property('cover')
    created_at = getset_property('created_at')  # datetime
    updated_at = getset_property('updated_at')  # datetime
    songs_count = getset_property('songs_count')
    creator = getset_property('creator')
    released_at = getset_property('released_at')  # str


class TableMetaWidget(MetaWidget):

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.cover_label = CoverLabelV2(self)
        # these three widgets are in right layout
        self.title_label = QLabel(self)
        self.meta_label = QLabel(self)
        # this spacer item is used as a stretch in right layout,
        # it's  width and height is not so important, we set them to 0
        self.text_spacer = QSpacerItem(0, 0, QSizePolicy.Minimum, QSizePolicy.Expanding)

        # self.title_label.setTextFormat(Qt.RichText)
        self.meta_label.setTextFormat(Qt.RichText)

        self._setup_ui()
        self._refresh()

    def _setup_ui(self):
        font = self.font()
        font.setPixelSize(20)
        font.setWeight(QFont.DemiBold)

        self.cover_label.setMinimumWidth(150)
        self.cover_label.setMaximumWidth(200)
        self.title_label.setFont(font)

        self._v_layout = QVBoxLayout(self)
        self._h_layout = QHBoxLayout()
        self._right_layout = QVBoxLayout()
        self._right_layout.addStretch(0)
        self._right_layout.addWidget(self.title_label)
        self._right_layout.addWidget(self.meta_label)
        self._h_layout.addWidget(self.cover_label)
        self._h_layout.setAlignment(self.cover_label, Qt.AlignTop)
        self._h_layout.addLayout(self._right_layout)
        self._h_layout.setStretchFactor(self._right_layout, 2)
        self._h_layout.setStretchFactor(self.cover_label, 1)
        self._h_layout.setContentsMargins(0, 30, 0, 0)
        self._h_layout.setSpacing(30)
        self._v_layout.addLayout(self._h_layout)

        self._right_layout.setContentsMargins(0, 0, 0, 0)
        self._right_layout.setSpacing(5)

        # left margin is same as toolbar left margin
        self.layout().setContentsMargins(0, 0, 30, 0)
        self.layout().setSpacing(0)

    def add_tabbar(self, tabbar):
        self._right_layout.addWidget(tabbar)
        self._right_layout.setAlignment(tabbar, Qt.AlignLeft)

    def set_cover_image(self, image):
        if image is not None:
            self.cover_label.show()
            self._right_layout.addItem(self.text_spacer)
        self.cover_label.show_img(image)
        self.updateGeometry()

    def on_property_updated(self, name):
        if name in ('created_at', 'updated_at', 'songs_count', 'creator',
                    'source', 'released_at'):
            self._refresh_meta_label()
        elif name in ('title', 'subtitle'):
            self._refresh_title()
        elif name == 'cover':
            self._refresh_cover()

    def _refresh_meta_label(self):
        creator = self.creator
        # icon: 👤
        creator_part = creator if creator else ''
        released_part = created_part = updated_part = songs_count_part = source_part = ''
        if self.source:
            source_part = f'<code style="color: gray;">{self.source}</code>'
        if self.updated_at:
            updated_part = '🕒 更新于 <code style="font-size: small">{}</code>'\
                .format(self.updated_at.strftime('%Y-%m-%d'))
        if self.created_at:
            created_part = '🕛 创建于 <code style="font-size: small">{}</code>'\
                .format(self.created_at.strftime('%Y-%m-%d'))
        if self.released_at:
            released_part = f'🕛 发布于 <code style="font-size: small">{self.released_at}</code>'  # noqa
        if self.songs_count is not None:
            text = self.songs_count if self.songs_count != -1 else '未知'
            songs_count_part = f'<code style="font-size: small">{text}</code> 首歌曲'
        parts = [
            creator_part,
            created_part,
            updated_part,
            songs_count_part,
            released_part,
            source_part,
        ]
        valid_parts = [p for p in parts if p]
        if valid_parts:
            content = ' • '.join(valid_parts)
            text = f'<span>{content}</span>'
            # TODO: add linkActivated callback for meta_label
            self.meta_label.setText(text)
            self.meta_label.show()
        else:
            self.meta_label.hide()

    def _refresh_cover(self):
        if not self.cover:
            self.cover_label.hide()
            self._right_layout.removeItem(self.text_spacer)
        else:
            self._right_layout.addItem(self.text_spacer)
        self.updateGeometry()

    def _refresh_title(self):
        if self.title:
            self.title_label.show()
            self.title_label.setToolTip(self.title)
            # Please refresh when the widget is resized.
            title = elided_text(self.title,
                                self.title_label.width(),
                                self.title_label.font())
            self.title_label.setText(title)
        else:
            self.title_label.hide()

    def _refresh(self):
        self._refresh_title()
        self._refresh_meta_label()
        self._refresh_cover()

    def sizeHint(self):
        super_size = super().sizeHint()
        if self.cover_label.isVisible():
            margins = self.layout().contentsMargins()
            v_margin = margins.top() + margins.bottom()
            height = self.cover_label.sizeHint().height() + v_margin
            return QSize(super_size.width(), height)
        return super_size

    def resizeEvent(self, e):
        super().resizeEvent(e)
        self._refresh_title()
        # HELP: think about a more elegant way
        # Currently, right panel draw background image on meta widget
        # and bottom panel, when meta widget is resized, the background
        # image will also be scaled, so we need to repaint on bottom panel
        # and meta widget. However, by default, qt will only repaint
        # meta widget in this case, so we trigger bottom panel update manually
        #
        # type ignore: parent should be TableContainer.
        self.parent()._app.ui.bottom_panel.update()  # type: ignore[attr-defined]



from collections import defaultdict
from difflib import SequenceMatcher
from typing import Any

from feeluown.library import resolve, reverse
from .base import AbstractHandler


def score(src, tar):
    s1 = int(SequenceMatcher(None, src, tar).ratio() * 50)
    s2 = 50
    delimiters = [' - ', '-', ' ']
    parts = []
    for delimiter in delimiters:
        if delimiter in src:
            parts = src.split(delimiter)
            break
    if parts:
        each_part_s = int(50 / len(parts))
        for part in parts:
            if part not in tar:
                s2 -= each_part_s
    return s1 + s2


def repr_song(song):
    return song.title + song.artists_name + song.album_name


class PlayerHandler(AbstractHandler):
    cmds = ('play', 'pause', 'stop', 'resume', 'toggle', )

    def handle(self, cmd):  # pylint: disable=inconsistent-return-statements
        # pylint: disable=no-else-return
        # TODO: 支持设置是否显示视频
        if cmd.action == 'play':
            s = ' '.join(cmd.args)
            return self.play(s)
        elif cmd.action == 'pause':
            # FIXME: please follow ``Law of Demeter``
            self.player.pause()
        elif cmd.action == 'stop':
            self.player.stop()
        elif cmd.action == 'resume':
            self.player.resume()
        elif cmd.action == 'toggle':
            self.player.toggle()

    def play(self, s):  # pylint: disable=inconsistent-return-statements
        # pylint: disable=no-else-return
        if s.startswith('fuo://'):
            model = resolve(s)
            if model is None:
                return 'Invalid fuo uri.'
            self._app.playlist.play_model(model)
            return
        elif s.startswith('http'):
            return self.player.play(s, video=False)

        # 取每个提供方的第一个搜索结果
        source_song_map: Any = defaultdict()
        for result in self.library.search(s):
            for song in result.songs:
                if song.source in source_song_map:
                    break
                source_song_map[song.source] = song
        songs = list(source_song_map.values())

        if songs:
            songs = sorted(songs, key=lambda song: score(s, repr_song(song)),
                           reverse=True)
            msg = f'select:\t{reverse(songs[0], as_line=True)}\n'
            self.playlist.play_model(songs[0])
            lines = []
            for song in songs[1:]:
                lines.append('\t' + reverse(song, as_line=True))
            if lines:
                msg += 'options::\n' + '\n'.join(lines)
            return msg
        else:
            return 'No song has been found.'



import sys
import os
import feeluown.__main__


if __name__ == '__main__':
    if hasattr(sys, '_MEIPASS'):
        os.chdir(sys._MEIPASS)

    os.environ.setdefault('FUO_LOG_TO_FILE', '1')
    feeluown.__main__.main()



import logging
import random

from PyQt5.QtCore import (
    pyqtSignal, Qt, QSize, QRect, QRectF,
    QAbstractListModel, QModelIndex,
)
from PyQt5.QtGui import (
    QPainter, QPixmap, QImage, QColor, QPalette, QBrush,
    QFontMetrics, QTextOption,
)
from PyQt5.QtWidgets import (
    QFrame, QListView, QStyle, QStyledItemDelegate
)

from feeluown.utils import aio
from feeluown.library import reverse
from feeluown.gui.helpers import (
    ItemViewNoScrollMixin, ReaderFetchMoreMixin, resize_font, SOLARIZED_COLORS
)


logger = logging.getLogger(__name__)
Fetching = object()


class BaseSongMiniCardListModel(QAbstractListModel):
    def __init__(self, fetch_image, parent=None):
        super().__init__(parent)

        self._items = []
        self.fetch_image = fetch_image
        self.pixmaps = {}  # {uri: (Option<pixmap>, Option<color>)}

    def rowCount(self, _=QModelIndex()):
        return len(self._items)

    def flags(self, index):
        if not index.isValid():
            return 0
        flags = Qt.ItemIsSelectable | Qt.ItemIsEnabled
        return flags

    def _fetch_image_callback(self, item):
        # TODO: duplicate code with ImgListModel
        def cb(content):
            uri = reverse(item)
            if content is None:
                self.pixmaps[uri] = (self.pixmaps[uri][1], None)
                return

            img = QImage()
            img.loadFromData(content)
            pixmap = QPixmap(img)
            self.pixmaps[uri] = (pixmap, None)
            row = self._items.index(item)
            top_left = self.createIndex(row, 0)
            bottom_right = self.createIndex(row, 0)
            self.dataChanged.emit(top_left, bottom_right)
        return cb

    def get_pixmap_unblocking(self, song):
        """
        return QColor means the song has no pixmap or the pixmap is currently not feched.
        """
        uri = reverse(song)
        if uri in self.pixmaps:
            pixmap, color = self.pixmaps[uri]
            if pixmap is Fetching:
                return color
            return pixmap
        aio.run_afn(self.fetch_image, song, self._fetch_image_callback(song))
        color = QColor(random.choice(list(SOLARIZED_COLORS.values())))
        color.setAlphaF(0.8)
        self.pixmaps[uri] = (Fetching, color)
        return color

    def data(self, index, role=Qt.DisplayRole):
        offset = index.row()
        if not index.isValid() or offset >= len(self._items):
            return None
        if role == Qt.DisplayRole:
            return self._items[offset].title_display
        elif role == Qt.UserRole:
            song = self._items[offset]
            pixmap = self.get_pixmap_unblocking(song)
            return (song, pixmap)
        return None


class SongMiniCardListModel(BaseSongMiniCardListModel, ReaderFetchMoreMixin):
    def __init__(self, reader, fetch_image, parent=None):
        super().__init__(fetch_image, parent)

        self._reader = reader
        self._fetch_more_step = 10
        self._is_fetching = False


class SongMiniCardListDelegate(QStyledItemDelegate):
    img_padding = 2

    def __init__(
        self,
        view,
        card_min_width=200,
        card_height=40,
        card_right_spacing=10,
        card_padding=(3, 3, 3, 0),
        hover_color_role=QPalette.Window,
    ):
        """
        QListView.setSpacing set spacing around the item, however, sometimes
        the left spacing is unneeded. `card_left_padding` is used to customize
        the behaviour.
        """
        super().__init__(parent=view)

        self.view = view
        self.card_min_width = card_min_width
        self.card_height = card_height
        self.card_right_spacing = card_right_spacing
        self.hover_color_role = hover_color_role
        self.card_top_padding = card_padding[1]
        self.card_bottom_padding = card_padding[3]
        self.card_left_padding = card_padding[0]

    def item_sizehint(self) -> tuple:
        # HELP: listview needs about 20 spacing left on macOS
        width = max(self.view.width() - 20, self.card_min_width)

        # according to our algorithm, when the widget width is:
        #   2(card_min_width + card_right_spacing) + card_right_spacing - 1,
        # the card width can take the maximum width, it will be:
        #   CardMaxWidth = 2 * card_min_width + card_right_spacing - 1

        # calculate max column count
        count = (width - self.card_right_spacing) // \
            (self.card_min_width + self.card_right_spacing)
        count = max(count, 1)
        item_width = (width - ((count + 1) * self.card_right_spacing)) // count
        return (item_width,
                self.card_height + self.card_top_padding + self.card_bottom_padding)

    def paint(self, painter, option, index):
        card_top_padding = self.card_top_padding
        card_right_spacing = self.card_right_spacing
        card_height = self.card_height
        card_left_padding = self.card_left_padding
        border_radius = 3

        rect = option.rect
        # HACK(cosven): from the QFontMetrics doc, there is usually a small spacing
        # between a character and the font rect highest/lowest position.
        # Assume the spacing is 2px here.
        img_padding = self.img_padding
        cover_height = card_height - 2 * img_padding
        cover_width = cover_height
        song, obj = index.data(Qt.UserRole)
        if obj is None:
            return

        selected = option.state & QStyle.State_Selected
        if selected:
            painter.save()
            painter.setPen(Qt.NoPen)
            painter.setBrush(option.palette.color(QPalette.Highlight))
            painter.drawRect(rect)
            painter.restore()
        elif option.state & QStyle.State_MouseOver:
            painter.save()
            painter.setPen(Qt.NoPen)
            painter.setBrush(option.palette.color(self.hover_color_role))
            painter.drawRect(rect)
            painter.restore()

        painter.save()
        painter.translate(rect.x() + card_left_padding,  rect.y() + card_top_padding)

        if selected:
            text_color = option.palette.color(QPalette.HighlightedText)
            non_text_color = QColor(text_color)
            non_text_color.setAlpha(200)
        else:
            text_color = option.palette.color(QPalette.Text)
            if text_color.lightness() > 150:
                non_text_color = text_color.darker(140)
            else:
                non_text_color = text_color.lighter(150)
            non_text_color.setAlpha(100)

        # Draw image.
        painter.save()
        painter.translate(0, img_padding)
        self.paint_pixmap(painter, non_text_color, obj,
                          cover_width, cover_height, border_radius)
        painter.restore()

        # Draw text.
        painter.save()
        text_width = rect.width() - cover_width - \
            card_left_padding * 2 - card_right_spacing
        painter.translate(cover_width + card_left_padding, 0)
        title = index.data(Qt.DisplayRole)
        subtitle = f'{song.artists_name_display} • {song.album_name_display}'
        # Note this is not a bool object.
        is_enabled = option.state & QStyle.State_Enabled
        self.paint_text(
            painter, is_enabled, title, subtitle, text_color, non_text_color,
            text_width, card_height
        )
        painter.restore()

        painter.restore()

    def paint_text(self, painter, is_enabled, title, subtitle,
                   text_color, non_text_color, text_width, text_height):
        each_height = text_height//2
        text_option = QTextOption()
        text_option.setAlignment(Qt.AlignLeft | Qt.AlignVCenter)

        # Draw title.
        title_rect = QRectF(0, 0, text_width, each_height)
        fm = QFontMetrics(painter.font())
        if is_enabled:
            painter.setPen(text_color)
        else:
            painter.setPen(non_text_color)
        elided_title = fm.elidedText(title, Qt.ElideRight, int(text_width))
        painter.drawText(title_rect, elided_title, text_option)

        # Draw subtitle.
        painter.translate(0, each_height)
        subtitle_rect = QRectF(0, 0, text_width, each_height)
        elided_title = fm.elidedText(subtitle, Qt.ElideRight, int(text_width))
        font = painter.font()
        resize_font(font, -1)
        painter.setFont(font)
        fm = QFontMetrics(font)
        painter.setPen(non_text_color)
        painter.drawText(subtitle_rect, elided_title, text_option)

    def paint_pixmap(self, painter, border_color, decoration,
                     width, height, border_radius):
        painter.setRenderHint(QPainter.Antialiasing)
        pen = painter.pen()
        pen.setColor(border_color)
        painter.setPen(pen)
        if isinstance(decoration, QColor):
            color = decoration
            brush = QBrush(color)
            painter.setBrush(brush)
        else:
            if decoration.height() < decoration.width():
                pixmap = decoration.scaledToHeight(height, Qt.SmoothTransformation)
            else:
                pixmap = decoration.scaledToWidth(width, Qt.SmoothTransformation)
            brush = QBrush(pixmap)
            painter.setBrush(brush)
        cover_rect = QRect(0, 0, width, height)
        painter.drawRoundedRect(cover_rect, border_radius, border_radius)

    def sizeHint(self, option, index):
        size = super().sizeHint(option, index)
        if index.isValid():
            return QSize(*self.item_sizehint())
        return size


class SongMiniCardListView(ItemViewNoScrollMixin, QListView):

    play_song_needed = pyqtSignal([object])

    def __init__(self, parent=None, **kwargs):
        super().__init__(parent=parent, **kwargs)

        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)
        self.setMouseTracking(True)
        self.setFrameShape(QFrame.NoFrame)
        self.setViewMode(QListView.IconMode)
        self.setResizeMode(QListView.Adjust)
        self.setWrapping(True)
        self.initialize()

        self.activated.connect(self._on_activated)

    def _on_activated(self, index):
        self.play_song_needed.emit(index.data(Qt.UserRole)[0])



import sys
import logging


logger = logging.getLogger(__name__)


async def run_nowplaying_server(app):
    try:
        await run_nowplaying_server_internal(app)
    except ImportError as e:
        logger.warning(f"run nowplaying service failed: '{e}'")
    except:  # noqa
        logger.exception('run nowplaying service error')


def get_service_cls():
    """
    Only public for testing.
    """
    # pylint: disable=import-outside-toplevel
    from .nowplaying import NowPlayingService

    if sys.platform == 'darwin':
        from .macos import MacosMixin

        class Service(MacosMixin, NowPlayingService):
            pass
    else:
        Service = NowPlayingService
    return Service


async def run_nowplaying_server_internal(app):
    # pylint: disable=import-outside-toplevel
    if sys.platform == 'linux':
        from .linux import run_mpris2_server
        run_mpris2_server(app)
    elif sys.platform in ('win32', 'darwin'):
        service = get_service_cls()(app)
        await service.start()
    else:
        logger.warning('nowplaying is not supported on %s', sys.platform)



import math
from typing import Optional

from PyQt5.QtCore import Qt, QRect, QPoint, QPointF
from PyQt5.QtGui import QPainter, QBrush, QPixmap, QImage, QColor, QPolygonF, QPalette
from PyQt5.QtWidgets import QWidget

from feeluown.gui.helpers import random_solarized_color, painter_save


class PixmapDrawer:
    """Draw pixmap on a widget with radius.

    The pixmap will be scaled to the width of the widget.
    """
    def __init__(self, img, widget: QWidget, radius: int = 0):
        """
        :param widget: a object which has width() and height() method.
        """
        self._widget_last_width = widget.width()
        self._widget = widget
        self._radius = radius

        if img is None:
            self._color = random_solarized_color()
            self._img = None
            self._pixmap = None
        else:
            self._img = img
            self._color = None
            new_img = img.scaledToWidth(self._widget_last_width, Qt.SmoothTransformation)
            self._pixmap = QPixmap(new_img)

    @classmethod
    def from_img_data(cls, img_data, *args, **kwargs):
        img = QImage()
        img.loadFromData(img_data)
        return cls(img, *args, **kwargs)

    def get_img(self) -> Optional[QImage]:
        return self._img

    def get_pixmap(self) -> Optional[QPixmap]:
        return self._pixmap

    def maybe_update_pixmap(self):
        if self._widget.width() != self._widget_last_width:
            self._widget_last_width = self._widget.width()
            assert self._img is not None
            new_img = self._img.scaledToWidth(self._widget_last_width,
                                              Qt.SmoothTransformation)
            self._pixmap = QPixmap(new_img)

    def draw(self, painter):
        painter.save()
        painter.setRenderHint(QPainter.Antialiasing)
        painter.setRenderHint(QPainter.SmoothPixmapTransform)
        if self._pixmap is None:
            self._draw_random_color(painter)
        else:
            self._draw_pixmap(painter)
        painter.restore()

    def _get_radius(self):
        return self._radius if self._radius >= 1 else self._widget.width() * self._radius

    def _draw_random_color(self, painter: QPainter):
        brush = QBrush(self._color)
        painter.setBrush(brush)
        painter.setPen(Qt.NoPen)
        rect = self._widget.rect()
        if self._radius == 0:
            painter.drawRect(rect)
        else:
            radius = self._get_radius()
            painter.drawRoundedRect(rect, radius, radius)

    def _draw_pixmap(self, painter: QPainter):
        assert self._pixmap is not None

        self.maybe_update_pixmap()
        brush = QBrush(self._pixmap)
        painter.setBrush(brush)
        painter.setPen(Qt.NoPen)
        radius = self._radius
        size = self._pixmap.size()
        y = (size.height() - self._widget.height()) // 2
        painter.save()
        painter.translate(0, -y)
        rect = QRect(0, 0, self._widget.width(), size.height())
        if radius == 0:
            painter.drawRect(rect)
        else:
            radius = radius if self._radius >= 1 else self._widget.width() * self._radius
            painter.drawRoundedRect(rect, radius, radius)
        painter.restore()


class AvatarIconDrawer:
    def __init__(self, length, padding, fg_color=None):
        self._length = length
        self._padding = padding

        self.fg_color = fg_color

    def draw(self, painter):
        pen = painter.pen()
        pen.setWidthF(1.5)
        painter.setPen(pen)

        if self.fg_color:
            painter.setPen(QColor(self.fg_color))

        diameter = self._length // 3
        # Draw circle.
        painter.drawEllipse(diameter, self._padding, diameter, diameter)
        # Draw body.
        x, y = self._padding, self._length // 2
        width, height = self._length // 2, self._length // 2
        painter.drawArc(x, y, width, height, 0, 60*16)
        painter.drawArc(x, y, width, height, 120*16, 60*16)


class PlusIconDrawer:
    def __init__(self, length, padding):
        self.top = QPoint(length//2, padding)
        self.bottom = QPoint(length//2, length - padding)
        self.left = QPoint(padding, length//2)
        self.right = QPoint(length-padding, length//2)

    def draw(self, painter):
        pen = painter.pen()
        pen.setWidthF(1.5)
        painter.setPen(pen)
        painter.drawLine(self.top, self.bottom)
        painter.drawLine(self.left, self.right)


class TriangleIconDrawer:
    def __init__(self, length, padding, direction='up', brush=False):
        self._length = length
        self._padding = padding
        self.brush = brush

        # initialized when direction is set
        # polygon points are orgonized in anticlockwise direction
        self.triangle = QPolygonF([])
        self.set_direction(direction)

    def set_direction(self, direction):
        length = self._length
        padding = self._padding

        half = length / 2
        diameter = length - 2 * padding

        real_padding = (length - diameter) / 2
        center_top = QPointF(half, real_padding)
        center_bottom = QPointF(half, real_padding + diameter)
        left = QPointF(real_padding, half)
        right = QPointF(real_padding + diameter, half)

        d60 = diameter / 2 * 0.87  # sin60
        d30 = diameter / 2 / 2    # sin30

        if direction in ('left', 'right'):
            left_x = half - d30
            top_y = half - d60
            bottom_y = half + d60
            right_x = half + d30
        else:
            left_x = half - d60
            top_y = half - d30
            bottom_y = half + d30
            right_x = half + d60

        left_top = QPointF(left_x, top_y)
        left_bottom = QPointF(left_x, bottom_y)
        right_top = QPointF(right_x, top_y)
        right_bottom = QPointF(right_x, bottom_y)

        if direction == 'up':
            self.triangle = QPolygonF([center_top, left_bottom, right_bottom])
        elif direction == 'down':
            self.triangle = QPolygonF([center_bottom, right_top, left_top])
        elif direction == 'left':
            self.triangle = QPolygonF([left, right_bottom, right_top])
        elif direction == 'right':
            self.triangle = QPolygonF([right, left_top, left_bottom])
        else:
            raise ValueError('direction must be one of up/down/left/right')

    def draw(self, painter):
        pen = painter.pen()
        pen.setWidthF(1.5)
        painter.setPen(pen)
        with painter_save(painter):
            if self.brush is True:
                painter.setBrush(pen.color())
            painter.drawPolygon(self.triangle)


class HomeIconDrawer:
    def __init__(self, length, padding):
        icon_length = length
        diff = 1  # root/body width diff
        h_padding = v_padding = padding

        body_left_x = h_padding + diff*2
        body_right_x = icon_length - h_padding - diff*2
        body_top_x = icon_length // 2

        self._roof = QPoint(icon_length // 2, v_padding)
        self._root_left = QPoint(h_padding, icon_length // 2 + diff)
        self._root_right = QPoint(icon_length - h_padding, icon_length // 2 + diff)

        self._body_bottom_left = QPoint(body_left_x, icon_length - v_padding)
        self._body_bottom_right = QPoint(body_right_x, icon_length - v_padding)
        self._body_top_left = QPoint(body_left_x, body_top_x)
        self._body_top_right = QPoint(body_right_x, body_top_x)

    def paint(self, painter):
        pen = painter.pen()
        pen.setWidthF(1.5)
        painter.setPen(pen)
        painter.drawLine(self._roof, self._root_left)
        painter.drawLine(self._roof, self._root_right)
        painter.drawLine(self._body_bottom_left, self._body_bottom_right)
        painter.drawLine(self._body_top_left, self._body_bottom_left)
        painter.drawLine(self._body_top_right, self._body_bottom_right)


class CalendarIconDrawer:
    def __init__(self, length, padding):
        self._body_x = self._body_y = padding
        self._body_width = length - 2 * padding
        self._radius = 3
        self._h_line_y = self._body_y + self._body_width // 4

    def paint(self, painter: QPainter):
        pen = painter.pen()
        pen.setWidthF(1.5)
        painter.setPen(pen)
        body_rect = QRect(self._body_x, self._body_x, self._body_width, self._body_width)
        painter.drawRoundedRect(body_rect, self._radius, self._radius)
        painter.drawLine(QPoint(self._body_x, self._h_line_y),
                         QPoint(self._body_x + self._body_width, self._h_line_y))


class RankIconDrawer:
    def __init__(self, length, padding):
        body = length - 2*padding
        body_2 = body // 2
        body_8 = body // 8
        body_3 = body // 3
        _top_right_x = length - padding
        _top_right_y = padding + body_8
        _bottom_left_y = padding + body - body_8

        self.p1 = QPoint(padding, _bottom_left_y)
        self.p2 = QPoint(padding + body_3, padding + body_3)
        self.p3 = QPoint(padding + body_2, padding + body_3 * 2)
        self.p4 = QPoint(_top_right_x, _top_right_y)
        self.p5 = QPoint(_top_right_x - body_3, _top_right_y)
        self.p6 = QPoint(_top_right_x, _top_right_y + body_3)

    def paint(self, painter: QPainter):
        pen = painter.pen()
        pen.setWidthF(1.5)
        painter.setPen(pen)
        painter.drawLine(self.p1, self.p2)
        painter.drawLine(self.p2, self.p3)
        painter.drawLine(self.p3, self.p4)
        painter.drawLine(self.p4, self.p5)
        painter.drawLine(self.p4, self.p6)


class StarIconDrawer:
    def __init__(self, length, padding):

        radius_outer = (length - 2*padding)//2
        length_half = length // 2
        radius_inner = radius_outer // 2
        center = QPointF(length_half, length_half)
        angle = math.pi / 2

        self._star_polygon = QPolygonF()
        for _ in range(5):
            outer_point = center + QPointF(
                radius_outer * math.cos(angle),
                -radius_outer * math.sin(angle)
            )
            self._star_polygon.append(outer_point)
            inner_point = center + QPointF(
                radius_inner * math.cos(angle + math.pi/5),
                -radius_inner * math.sin(angle + math.pi/5)
            )
            self._star_polygon.append(inner_point)
            angle += 2 * math.pi / 5

    def paint(self, painter: QPainter):
        pen = painter.pen()
        pen.setWidthF(1.5)
        painter.setPen(pen)
        painter.drawPolygon(self._star_polygon)


class VolumeIconDrawer:
    def __init__(self, length, padding):
        self._polygon = QPolygonF()
        self._padding = padding
        self._volume = 100

        body_length = length - 2 * padding
        body_length_4_1 = 0.25 * body_length
        body_length_2_1 = 0.5 * body_length
        body_length_4_3 = 0.75 * body_length
        body_length_3_1 = 0.33 * body_length
        body_length_3_2 = 0.66 * body_length

        left_top = QPointF(0, body_length_4_1)
        left_bottom = QPointF(0, body_length_4_3)
        mid_top = QPointF(body_length_3_1, body_length_4_1)
        mid_bottom = QPointF(body_length_3_1, body_length_4_3)
        right_top = QPointF(body_length_3_2, 0)
        right_bottom = QPointF(body_length_3_2, body_length)

        for p in (left_top, left_bottom, mid_bottom, right_bottom,
                  right_top, mid_top):
            self._polygon.append(p)

        line_right_top_y = 0.1 * body_length
        line_right_bottom_y = 0.9 * body_length
        line_left_top_y = 0.2 * body_length
        line_left_bottom_y = 0.8 * body_length
        line_left_x = 0.85 * body_length
        self._line1 = (QPointF(line_left_x, line_left_top_y),
                       QPointF(body_length, line_right_top_y))
        self._line2 = (QPointF(line_left_x, body_length_2_1),
                       QPointF(body_length, body_length_2_1))
        self._line3 = (QPointF(line_left_x, line_left_bottom_y),
                       QPointF(body_length, line_right_bottom_y))

    def set_volume(self, volume):
        self._volume = volume

    def draw(self, painter: QPainter, palette: QPalette):
        pen = painter.pen()
        pen.setWidthF(1.5)
        painter.setPen(pen)
        with painter_save(painter):
            painter.translate(self._padding, self._padding)
            painter.drawPolygon(self._polygon)

            if self._volume >= 66:
                lines = (self._line1, self._line2, self._line3)
                disabled_lines = ()
            elif self._volume >= 33:
                lines = (self._line2, self._line3)
                disabled_lines = (self._line1, )
            elif self._volume > 0:
                lines = (self._line3, )
                disabled_lines = (self._line1, self._line2)
            else:
                lines = ()
                disabled_lines = (self._line1, self._line2, self._line3)
            for line in lines:
                painter.drawLine(*line)
            pen = painter.pen()
            pen.setColor(palette.color(QPalette.Disabled, QPalette.ButtonText))
            painter.setPen(pen)
            for line in disabled_lines:
                painter.drawLine(*line)
